Package: ambiorix
Title: Web Framework Inspired by 'Express.js'
Version: 2.2.0
Authors@R: 
  c(
    person(given = "John",
           family = "Coene",
           role = c("aut", "cre"),
           email = "jcoenep@gmail.com",
           comment = c(ORCID = "0000-0002-6637-4107")),
    person(family = "Opifex", role = c("fnd")),
    person(given = "Kennedy", family = "Mwavu", role = "ctb", comment = c(ORCID = "0009-0006-3157-7234"))
  )
Description: A web framework inspired by 'express.js' to build
  any web service from multi-page websites to 'RESTful' 
  application programming interfaces.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: fs, log, cli, glue, httpuv, methods, assertthat, webutils (>=
        1.2.0), yyjsonr (>= 0.1.20)
Suggests: mime, readr, readxl, ggplot2, promises, jsonlite, websocket,
        htmltools, commonmark, htmlwidgets, testthat (>= 3.0.0)
URL: https://github.com/ambiorix-web/ambiorix, https://ambiorix.dev
BugReports: https://github.com/ambiorix-web/ambiorix/issues
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2025-03-13 10:51:06 UTC; mwavu
Author: John Coene [aut, cre] (<https://orcid.org/0000-0002-6637-4107>),
  Opifex [fnd],
  Kennedy Mwavu [ctb] (<https://orcid.org/0009-0006-3157-7234>)
Maintainer: John Coene <jcoenep@gmail.com>
Repository: CRAN
Date/Publication: 2025-03-15 16:40:02 UTC
Built: R 4.3.3; ; 2025-03-18 02:10:56 UTC; unix

Package: ambiorix
Title: Web Framework Inspired by 'Express.js'
Version: 3.0.0
Authors@R: 
  c(
    person(given = "John",
           family = "Coene",
           role = "aut",
           email = "jcoenep@gmail.com",
           comment = c(ORCID = "0000-0002-6637-4107")),
    person(family = "Opifex", role = c("fnd")),
    person(given = "Kennedy", family = "Mwavu", role = "cre", email = "mwavukennedy@gmail.com", comment = c(ORCID = "0009-0006-3157-7234")),
    person(given = "Julio", family = "Collazos", role = "ctb", comment = c(ORCID = "0009-0006-5503-0997"))
  )
Description: A web framework inspired by 'express.js' to build
  any web service from multi-page websites to 'RESTful' 
  application programming interfaces.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: log, cli, glue, httpuv, methods, assertthat, webutils (>=
        1.2.0), yyjsonr (>= 0.1.20), mime (>= 0.13), promises (>=
        1.5.0)
Suggests: readr, readxl, ggplot2, jsonlite, websocket, htmltools,
        commonmark, htmlwidgets, testthat (>= 3.0.0)
URL: https://github.com/ambiorix-web/ambiorix, https://ambiorix.dev
BugReports: https://github.com/ambiorix-web/ambiorix/issues
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-25 03:44:23 UTC; mwavu
Author: John Coene [aut] (ORCID: <https://orcid.org/0000-0002-6637-4107>),
  Opifex [fnd],
  Kennedy Mwavu [cre] (ORCID: <https://orcid.org/0009-0006-3157-7234>),
  Julio Collazos [ctb] (ORCID: <https://orcid.org/0009-0006-5503-0997>)
Maintainer: Kennedy Mwavu <mwavukennedy@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-25 04:00:03 UTC
Built: R 4.6.0; ; 2026-04-26 05:31:27 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(convert_body,default)
S3method(convert_body,factor)
S3method(convert_body,shiny.tag)
S3method(convert_body,shiny.tag.list)
S3method(print,ambiorixResponse)
S3method(print,cookie)
S3method(print,cookieParser)
S3method(print,cookiePreprocessor)
S3method(print,forward)
S3method(print,jobj)
S3method(print,pathToPattern)
S3method(print,renderer)
S3method(print,responsePreHook)
S3method(print,robj)
export(Ambiorix)
export(Request)
export(Response)
export(Router)
export(Websocket)
export(as_cookie_parser)
export(as_cookie_preprocessor)
export(as_path_to_pattern)
export(as_renderer)
export(content_csv)
export(content_html)
export(content_json)
export(content_plain)
export(content_protobuf)
export(content_tsv)
export(copy_websocket_client)
export(default_cookie_parser)
export(forward)
export(import)
export(jobj)
export(mockRequest)
export(new_log)
export(parse_form_urlencoded)
export(parse_json)
export(parse_multipart)
export(pre_hook)
export(response)
export(response_404)
export(response_500)
export(robj)
export(serialise)
export(set_log_error)
export(set_log_info)
export(set_log_success)
export(stop_all)
export(token_create)
export(use_html_template)
importFrom(assertthat,assert_that)
importFrom(methods,formalArgs)
importFrom(utils,browseURL)

<div align="center">

# ambiorix

<img src="man/figures/ambiorix.png" height = "200px"/>

<!-- badges: start -->

[![R-CMD-check](https://github.com/ambiorix-web/ambiorix/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ambiorix-web/ambiorix/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/ambiorix)](https://CRAN.R-project.org/package=ambiorix)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)

<!-- badges: end -->

[Website](https://ambiorix.dev) | [CLI](https://github.com/devOpifex/ambiorix-cli) | [Generator](https://github.com/ambiorix-web/ambiorix.generator) | [Docker](https://hub.docker.com/r/jcoenep/ambiorix) | [Load Balancer](https://github.com/ambiorix-web/belgic)

Web framework for R based on [httpuv](https://github.com/rstudio/httpuv) and inspired by [express.js](https://github.com/expressjs/express).

</div>

## Install

The stable version is available on CRAN with:

```r
install.packages("ambiorix")
```

You can also install the development version from Github:

```r
remotes::install_github("ambiorix-web/ambiorix")
```

## Example

```r
library(ambiorix)

app <- Ambiorix$new(port = 3000L)

app$get("/", function(req, res) {
  res$send("Hello, World!")
})

app$get("/api/v1/users", function(req, res) {
  users <- data.frame(
    uid = 1:3,
    firstname = c("Alice", "Bob", "Cate"),
    lastname = c("Queen", "Jeremy", "Reece"),
    active = c(TRUE, FALSE, TRUE)
  )

  res$json(users)
})

app$start()
```

## Middlewares

- [druid](https://github.com/ambiorix-web/druid) Logger
- [alesia](https://github.com/ambiorix-web/alesia) Minifier
- [eburones](https://github.com/ambiorix-web/eburones) Sessions
- [agris](https://github.com/ambiorix-web/druid) Security
- [scilis](https://github.com/ambiorix-web/scilis) Cookies
- [titan](https://github.com/devOpifex/titan) Prometheus metrics
- [surf](https://github.com/ambiorix-web/surf) CSRF protection
- [signaculum](https://github.com/ambiorix-web/signaculum) favicon
- [pugger](https://github.com/ambiorix-web/pugger) Pug engine
- [jader](https://github.com/ambiorix-web/jader) Jade engine

## Tools & Extensions

- [belgic](https://github.com/ambiorix-web/belgic) Load balancer
- [packer](https://github.com/JohnCoene/packer) JavaScript
- [CLI](https://github.com/devOpifex/ambiorix-cli) for generator
- [Generator](https://github.com/ambiorix-web/ambiorix.generator) Project generator
- [Docker](https://hub.docker.com/r/jcoenep/ambiorix) Docker image

## Contributing

Please note that the ambiorix project is released with a [Contributor Code of Conduct](https://contributor-covenant.org/version/2/0/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms.

#' Parse multipart form data
#'
#' Parses multipart form data, including file uploads, and returns the parsed fields as a list.
#'
#' @param req The request object.
#' @param ... Additional parameters passed to the parser function.
#'
#' @details
#' If a field is a file upload it is returned as a named list with:
#' - `value`: Raw vector representing the file contents. You must
#'    process this further (eg. convert to data.frame). See the examples section.
#' - `content_disposition`: Typically "form-data", indicating how the content
#'    is meant to be handled.
#' - `content_type`: MIME type of the uploaded file (e.g., "image/png" or "application/pdf").
#' - `name`: Name of the form input field.
#' - `filename`: Original name of the uploaded file.
#'
#' If no body data, an empty list is returned.
#'
#' ### Overriding Default Parser
#'
#' By default, `parse_multipart()` uses [webutils::parse_http()] internally.
#' You can override this globally by setting the `AMBIORIX_MULTIPART_FORM_DATA_PARSER` option:
#'
#' ```r
#' options(AMBIORIX_MULTIPART_FORM_DATA_PARSER = my_custom_parser)
#' ```
#'
#' Your custom parser function must accept the following parameters:
#' 1. `body`: Raw vector containing the form data.
#' 2. `content_type`: The 'Content-Type' header of the request as defined by the client.
#' 3. `...`: Additional optional parameters.
#'
#' @examples
#' if (interactive()) {
#'   library(ambiorix)
#'   library(htmltools)
#'   library(readxl)
#'
#'   page_links <- function() {
#'     Map(
#'       f = function(href, label) {
#'         tags$a(href = href, label)
#'       },
#'       c("/", "/about", "/contact"),
#'       c("Home", "About", "Contact")
#'     )
#'   }
#'
#'   forms <- function() {
#'     form1 <- tags$form(
#'       action = "/url-form-encoded",
#'       method = "POST",
#'       enctype = "application/x-www-form-urlencoded",
#'       tags$h4("form-url-encoded:"),
#'       tags$label(`for` = "first_name", "First Name"),
#'       tags$input(id = "first_name", name = "first_name", value = "John"),
#'       tags$label(`for` = "last_name", "Last Name"),
#'       tags$input(id = "last_name", name = "last_name", value = "Coene"),
#'       tags$button(type = "submit", "Submit")
#'     )
#'
#'     form2 <- tags$form(
#'       action = "/multipart-form-data",
#'       method = "POST",
#'       enctype = "multipart/form-data",
#'       tags$h4("multipart/form-data:"),
#'       tags$label(`for` = "email", "Email"),
#'       tags$input(id = "email", name = "email", value = "john@mail.com"),
#'       tags$label(`for` = "framework", "Framework"),
#'       tags$input(id = "framework", name = "framework", value = "ambiorix"),
#'       tags$label(`for` = "file", "Upload CSV file"),
#'       tags$input(type = "file", id = "file", name = "file", accept = ".csv"),
#'       tags$label(`for` = "file2", "Upload xlsx file"),
#'       tags$input(type = "file", id = "file2", name = "file2", accept = ".xlsx"),
#'       tags$button(type = "submit", "Submit")
#'     )
#'
#'     tagList(form1, form2)
#'   }
#'
#'   home_get <- function(req, res) {
#'     html <- tagList(
#'       page_links(),
#'       tags$h3("hello, world!"),
#'       forms()
#'     )
#'
#'     res$send(html)
#'   }
#'
#'   home_post <- function(req, res) {
#'     body <- parse_json(req)
#'     # print(body)
#'
#'     response <- list(
#'       code = 200L,
#'       msg = "hello, world"
#'     )
#'     res$json(response)
#'   }
#'
#'   url_form_encoded_post <- function(req, res) {
#'     body <- parse_form_urlencoded(req)
#'     # print(body)
#'
#'     list_items <- lapply(
#'       X = names(body),
#'       FUN = function(nm) {
#'         tags$li(
#'           nm,
#'           ":",
#'           body[[nm]]
#'         )
#'       }
#'     )
#'     input_vals <- tags$ul(list_items)
#'
#'     html <- tagList(
#'       page_links(),
#'       tags$h3("Request processed"),
#'       input_vals
#'     )
#'
#'     res$send(html)
#'   }
#'
#'   multipart_form_data_post <- function(req, res) {
#'     body <- parse_multipart(req)
#'
#'     list_items <- lapply(
#'       X = names(body),
#'       FUN = function(nm) {
#'         field <- body[[nm]]
#'
#'         # if 'field' is a file, parse it & print on console:
#'         is_file <- "filename" %in% names(field)
#'         is_csv <- is_file && identical(field[["content_type"]], "text/csv")
#'         is_xlsx <- is_file &&
#'           identical(
#'             field[["content_type"]],
#'             "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
#'           )
#'
#'         if (is_file) {
#'           file_path <- tempfile()
#'           writeBin(object = field$value, con = file_path)
#'           on.exit(unlink(x = file_path))
#'         }
#'
#'         if (is_csv) {
#'           # print(read.csv(file = file_path))
#'         }
#'
#'         if (is_xlsx) {
#'           # print(readxl::read_xlsx(path = file_path))
#'         }
#'
#'         tags$li(
#'           nm,
#'           ":",
#'           if (is_file) "printed on console" else field
#'         )
#'       }
#'     )
#'     input_vals <- tags$ul(list_items)
#'
#'     html <- tagList(
#'       page_links(),
#'       tags$h3("Request processed"),
#'       input_vals
#'     )
#'
#'     res$send(html)
#'   }
#'
#'   about_get <- function(req, res) {
#'     html <- tagList(
#'       page_links(),
#'       tags$h3("About Us")
#'     )
#'     res$send(html)
#'   }
#'
#'   contact_get <- function(req, res) {
#'     html <- tagList(
#'       page_links(),
#'       tags$h3("Get In Touch!")
#'     )
#'     res$send(html)
#'   }
#'
#'   app <- Ambiorix$new(port = 5000L)
#'
#'   app$
#'     get("/", home_get)$
#'     post("/", home_post)$
#'     get("/about", about_get)$
#'     get("/contact", contact_get)$
#'     post("/url-form-encoded", url_form_encoded_post)$
#'     post("/multipart-form-data", multipart_form_data_post)
#'
#'   app$start()
#' }
#' @seealso [parse_form_urlencoded()], [parse_json()]
#' @return Named list.
#' @export
parse_multipart <- function(req, ...) {
  on.exit(req$rook.input$rewind())
  body <- req$rook.input$read()
  if (identical(body, raw())) {
    return(list())
  }

  default <- function(body, content_type, ...) {
    webutils::parse_http(
      body = body,
      content_type = content_type,
      ...
    )
  }

  parser <- getOption(
    x = "AMBIORIX_MULTIPART_FORM_DATA_PARSER",
    default = default
  )
  parsed <- parser(body = body, content_type = req$CONTENT_TYPE, ...)

  if (!identical(parser, default)) {
    return(parsed)
  }

  raw_to_char <- function(x) rawToChar(as.raw(x))

  values <- lapply(
    X = parsed,
    FUN = function(item) {
      # return files as is:
      is_file <- "filename" %in% names(item)
      if (is_file) {
        return(item)
      }

      raw_to_char(item[["value"]])
    }
  )

  values
}

#' Parse application/x-www-form-urlencoded data
#'
#' @description
#' This function parses `application/x-www-form-urlencoded` data, typically used in form submissions.
#'
#' @param req The request object.
#' @param ... Additional parameters passed to the parser function.
#'
#' @return A list of parsed form fields, with each key representing a form field name and each value
#' representing the form field's value.
#'
#' @details
#'
#' ### Overriding Default Parser
#'
#' By default, `parse_form_urlencoded()` uses [webutils::parse_http()].
#' You can override this globally by setting the `AMBIORIX_FORM_URLENCODED_PARSER` option:
#'
#' ```r
#' options(AMBIORIX_FORM_URLENCODED_PARSER = my_other_custom_parser)
#' ```
#'
#' Your custom parser function *MUST* accept the following parameters:
#' 1. `body`: Raw vector containing the form data.
#' 2. `...`: Additional optional parameters.
#'
#' @inherit parse_multipart examples
#' @seealso [parse_multipart()], [parse_json()]
#' @return Named list
#' @export
parse_form_urlencoded <- function(req, ...) {
  on.exit(req$rook.input$rewind())
  body <- req$rook.input$read()
  if (identical(body, raw())) {
    return(list())
  }

  default <- function(body, ...) {
    webutils::parse_http(
      body = body,
      content_type = "application/x-www-form-urlencoded",
      ...
    )
  }

  parser <- getOption(x = "AMBIORIX_FORM_URLENCODED_PARSER", default = default)

  parser(body = body, ...)
}

#' Parse application/json data
#'
#' @description
#' This function parses JSON data from the request body.
#'
#' @param req The request object.
#' @param ... Additional parameters passed to the parser function.
#'
#' @return An R object (e.g., list or data frame) parsed from the JSON data.
#'
#' @details
#'
#' ### Overriding Default Parser
#'
#' By default, `parse_json()` uses [yyjsonr::read_json_raw()] for JSON parsing.
#' You can override this globally by setting the `AMBIORIX_JSON_PARSER` option:
#'
#' ```r
#' my_json_parser <- function(body, ...) {
#'   txt <- rawToChar(body)
#'   jsonlite::fromJSON(txt, ...)
#' }
#' options(AMBIORIX_JSON_PARSER = my_json_parser)
#' ```
#'
#' Your custom parser *MUST* accept the following parameters:
#' 1. `body`: Raw vector containing the JSON data.
#' 2. `...`: Additional optional parameters.
#'
#' @inherit parse_multipart examples
#' @seealso [parse_multipart()], [parse_form_urlencoded()]
#' @return Named list
#' @export
parse_json <- function(req, ...) {
  on.exit(req$rook.input$rewind())
  body <- req$rook.input$read()
  if (identical(body, raw())) {
    return(list())
  }

  parser <- getOption(
    x = "AMBIORIX_JSON_PARSER",
    default = yyjsonr::read_json_raw
  )
  parser(body, ...)
}

test_that("parse_json works correctly", {
  # empty body
  req <- mockRequest()
  req$rook.input <- list(read = function() raw(), rewind = function() NULL)
  result <- parse_json(req)
  expect_equal(result, list())

  # JSON object
  json_data <- '{"name": "John", "age": 30, "active": true}'
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(json_data),
    rewind = function() NULL
  )
  result <- parse_json(req)
  expected <- list(name = "John", age = 30L, active = TRUE)
  expect_equal(result, expected)

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw('{"invalid": json}'),
    rewind = function() NULL
  )
  expect_error(parse_json(req))

  # JSON array
  json_array <- '[{"id": 1}, {"id": 2}]'
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(json_array),
    rewind = function() NULL
  )
  result <- parse_json(req)
  expected <- data.frame(id = 1:2)
  expect_equal(result, expected)

  # empty JSON object
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw("{}"),
    rewind = function() NULL
  )
  result <- parse_json(req)
  expected <- list()
  names(expected) <- character()
  expect_equal(result, expected)

  # empty JSON array
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw("[]"),
    rewind = function() NULL
  )
  result <- parse_json(req)
  expected <- list()
  expect_equal(result, expected)

  # custom parser override
  custom_parser <- function(body, ...) {
    list(custom = "parsed", raw_length = length(body))
  }
  old_option <- getOption("AMBIORIX_JSON_PARSER")
  options(AMBIORIX_JSON_PARSER = custom_parser)

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw('{"test": "data"}'),
    rewind = function() NULL
  )
  result <- parse_json(req)
  expect_equal(result$custom, "parsed")
  expect_equal(result$raw_length, nchar('{"test": "data"}'))

  # restore original option
  options(AMBIORIX_JSON_PARSER = old_option)
})

test_that("parse_form_urlencoded works correctly", {
  # empty body
  req <- mockRequest()
  req$rook.input <- list(read = function() raw(), rewind = function() NULL)
  result <- parse_form_urlencoded(req)
  expect_equal(result, list())

  # simple form data
  form_data <- "name=John&age=30&active=true"
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(form_data),
    rewind = function() NULL
  )
  result <- parse_form_urlencoded(req)
  expect_equal(result$name, "John")
  expect_equal(result$age, "30")
  expect_equal(result$active, "true")

  # malformed data
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw("invalid=form=data=here"),
    rewind = function() NULL
  )
  # webutils is quite tolerant, will parse first value:
  result <- parse_form_urlencoded(req)
  expected <- list(invalid = "form")
  expect_equal(result, expected)

  # URL-encoded special characters
  form_data <- "message=Hello%20World%21&email=user%40example.com"
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(form_data),
    rewind = function() NULL
  )
  result <- parse_form_urlencoded(req)
  expect_equal(result$message, "Hello World!")
  expect_equal(result$email, "user@example.com")

  # empty values
  form_data <- "empty=&filled=value&another="
  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(form_data),
    rewind = function() NULL
  )
  result <- parse_form_urlencoded(req)
  expect_equal(result$empty, NA_character_)
  expect_equal(result$filled, "value")
  expect_equal(result$another, NA_character_)

  # custom parser override
  custom_parser <- function(body, ...) {
    list(custom = "form_parsed", body_length = length(body))
  }
  old_option <- getOption("AMBIORIX_FORM_URLENCODED_PARSER")
  options(AMBIORIX_FORM_URLENCODED_PARSER = custom_parser)

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw("test=data"),
    rewind = function() NULL
  )
  result <- parse_form_urlencoded(req)
  expect_equal(result$custom, "form_parsed")
  expect_equal(result$body_length, nchar("test=data"))

  # restore original option
  options(AMBIORIX_FORM_URLENCODED_PARSER = old_option)
})

test_that("parse_multipart works correctly", {
  # empty body
  req <- mockRequest()
  req$rook.input <- list(read = function() raw(), rewind = function() NULL)
  req$CONTENT_TYPE <- "multipart/form-data; boundary=----WebKitFormBoundary"
  result <- parse_multipart(req)
  expect_equal(result, list())

  # simple multipart data with text fields
  boundary <- "----WebKitFormBoundary7MA4YWxkTrZu0gW"
  multipart_body <- paste0(
    "--",
    boundary,
    "\r\n",
    "Content-Disposition: form-data; name=\"username\"\r\n\r\n",
    "john_doe\r\n",
    "--",
    boundary,
    "\r\n",
    "Content-Disposition: form-data; name=\"email\"\r\n\r\n",
    "john@example.com\r\n",
    "--",
    boundary,
    "--\r\n"
  )

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(multipart_body),
    rewind = function() NULL
  )
  req$CONTENT_TYPE <- paste0("multipart/form-data; boundary=", boundary)
  result <- parse_multipart(req)

  expect_equal(result$username, "john_doe")
  expect_equal(result$email, "john@example.com")

  # multipart with file upload
  file_content <- "file content here"
  multipart_with_file <- paste0(
    "--",
    boundary,
    "\r\n",
    "Content-Disposition: form-data; name=\"document\"; filename=\"test.txt\"\r\n",
    "Content-Type: text/plain\r\n\r\n",
    file_content,
    "\r\n",
    "--",
    boundary,
    "\r\n",
    "Content-Disposition: form-data; name=\"description\"\r\n\r\n",
    "A test file\r\n",
    "--",
    boundary,
    "--\r\n"
  )

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw(multipart_with_file),
    rewind = function() NULL
  )
  req$CONTENT_TYPE <- paste0("multipart/form-data; boundary=", boundary)
  result <- parse_multipart(req)

  # file should retain its structure
  expect_true("filename" %in% names(result$document))
  expect_equal(result$document$filename, "test.txt")
  expect_equal(result$document$content_type, "text/plain")
  expect_equal(rawToChar(result$document$value), file_content)

  # text field should be converted to character
  expect_equal(result$description, "A test file")

  # custom parser override
  custom_parser <- function(body, content_type, ...) {
    list(custom = "multipart_parsed", content_type = content_type)
  }
  old_option <- getOption("AMBIORIX_MULTIPART_FORM_DATA_PARSER")
  options(AMBIORIX_MULTIPART_FORM_DATA_PARSER = custom_parser)

  req <- mockRequest()
  req$rook.input <- list(
    read = function() charToRaw("dummy"),
    rewind = function() NULL
  )
  req$CONTENT_TYPE <- "multipart/form-data; boundary=test"
  result <- parse_multipart(req)
  expect_equal(result$custom, "multipart_parsed")
  expect_equal(result$content_type, "multipart/form-data; boundary=test")

  # restore original option
  options(AMBIORIX_MULTIPART_FORM_DATA_PARSER = old_option)
})

#' Replace Partials
#'
#' Replaces partial tags `[! ... !]` with their content.
#'
#' @param content File content where partials should be
#' replaced.
#' @param dir Directory of file from which `content` originates.
#'
#' @keywords internal
#' @noRd
replace_partials <- function(content, dir) {
  content <- apply_replace_partial(content, dir)

  if (any(grepl("\\[! .* !\\]", content))) {
    content <- replace_partials(content, dir)
  }

  return(content)
}

#' Replace Partial
#'
#' Replaces partial on a single line.
#' Assumes we have a single partial on a single line.
#'
#' @param line Single content line.
#' @param dir Base directory.
#'
#' @keywords internal
#' @noRd
replace_partial <- function(line, dir) {
  if (length(line) > 1) {
    line <- apply_replace_partial(line, dir)
  }

  if (!grepl("\\[! .* !\\]", line)) {
    return(line)
  }

  # every line is enclosed in <p> tag coming from markdown
  # we remove this, it's safe to assume that is not wanted.
  line <- gsub(
    "<p>\\[\\! ?",
    "[!",
    line
  )
  line <- gsub(
    " ?\\!\\]</p>",
    "!]",
    line
  )

  path <- trimws(gsub("\\[!|!\\]", "", line))

  # construct new base directory
  new_dir <- dirname(path)
  new_dir <- file.path(dir, new_dir)

  # build new path based on new directory
  path <- basename(path)
  path <- file.path(new_dir, path)

  # read lines
  lines <- read_lines_cached(path)

  # add new directory
  pat <- sprintf("[! %s/", new_dir)
  gsub("\\[! ", pat, lines)
}

#' Get Directory
#'
#' Retrieve directory from a file.
#'
#' @param file File to retrieve directory from.
#'
#' @keywords internal
#' @noRd
get_dir <- function(file) {
  dirname(normalizePath(file))
}

#' Replace Partial Vectorised
#'
#' Vectorised version of `replace_partial()`.
#'
#' @inheritParams replace_partials
#'
#' @keywords internal
#' @noRd
apply_replace_partial <- function(content, dir) {
  unlist(
    unname(
      sapply(content, replace_partial, dir)
    )
  )
}

#' R Object
#'
#' Treats a data element rendered in a response (`res$render`) as
#' a data object and ultimately uses [dput()].
#'
#' For instance in a template, `x <- [% var %]` will not work with
#' `res$render(data=list(var = "hello"))` because this will be replace
#' like `x <- hello` (missing quote): breaking the template. Using `robj` one would
#' obtain `x <- "hello"`.
#'
#' @param obj R object to treat.
#' @return Object of class "robj".
#' @examples
#' robj(1:10)
#' @export
robj <- function(obj) {
  assert_that(not_missing(obj))

  # Supress warnings otherwise
  # NULL, NA, and the likes
  # raise messages
  suppressWarnings(
    structure(obj, class = c("robj", class(obj)))
  )
}

#' @export
print.robj <- function(x, ...) {
  cli::cli_alert_info("R object")
  class(x) <- class(x)[!class(x) %in% "robj"]
  dput(x)
}

#' JSON Object
#'
#' Serialises an object to JSON in `res$render`.
#'
#' @param obj Object to serialise.
#' @examples
#' if (interactive()) {
#'   l <- list(a = "hello", b = 2L, c = 3)
#'   jobj(l)
#' }
#' @return Object of class "jobj".
#' @export
jobj <- function(obj) {
  suppressWarnings(
    structure(obj, class = c("jobj", class(obj)))
  )
}

#' @export
print.jobj <- function(x, ...) {
  cli::cli_alert_info("JSON object")
  serialise(x, ...)
}

#' Pre Hook Response
#'
#' @param content File content, a character vector.
#' @param data A list of data passed to `glue::glue_data`.
#'
#' @examples
#' my_prh <- function(self, content, data, ext, ...) {
#'   data$title <- "Mansion"
#'   pre_hook(content, data)
#' }
#'
#' #' Handler for GET at '/'
#' #'
#' #' @details Renders the homepage
#' #' @export
#' home_get <- function(req, res) {
#'   res$pre_render_hook(my_prh)
#'   res$render(
#'     file = "page.html",
#'     data = list(
#'       title = "Home"
#'     )
#'   )
#' }
#' @return A response pre-hook.
#' @export
pre_hook <- function(
  content,
  data
) {
  structure(
    list(
      content = content,
      data = data
    ),
    class = c("list", "responsePreHook")
  )
}

#' @keywords internal
is_pre_hook <- function(obj) {
  inherits(obj, "responsePreHook")
}

#' @export
print.responsePreHook <- function(x, ...) {
  cli::cli_alert_info("A response pre hook")
}

#' HTML Template
#'
#' Use [htmltools::htmlTemplate()] as renderer.
#' Passed to `use` method.
#'
#' @return A renderer function.
#' @examples
#' use_html_template()
#' @export
use_html_template <- function() {
  as_renderer(function(file, data) {
    if (tools::file_ext(file) != "html") {
      return()
    }

    data$filename <- file
    x <- do.call(
      htmltools::htmlTemplate,
      data
    )
    as.character(x)
  })
}

#' Render Tags
#'
#' @param lines Output of [read_lines()]
#' @param data Data to render, a `list`.
#'
#' @keywords internal
#' @noRd
render_tags <- function(lines, data) {
  new_lines <- c()
  n <- 0L
  str <- ""

  for (i in seq_along(lines)) {
    line <- lines[i]
    if (!grepl("\\[%|%\\]", line) && n == 0L) {
      new_lines <- c(new_lines, line)
      next()
    }

    if (!grepl("\\[%|%\\]", line) && n > 0L) {
      str <- paste(str, line)
      next()
    }

    opened <- lengths(regmatches(line, gregexpr("\\[%", line)))
    closed <- lengths(regmatches(line, gregexpr("%\\]", line)))

    n <- n + opened - closed

    str <- paste(str, line)

    # all in one line we render and continue
    if (n == 0L) {
      new_line <- glue::glue_data(data, str, .open = "[%", .close = "%]")
      new_lines <- c(new_lines, new_line)
      str <- ""
      next
    }
  }

  if (str != "") {
    message("error")
  }

  new_lines
}

#' Render HTML
#'
#' Evaluates a string to collect [htmltools::tags], evaluates,
#' and returns the render HTML as a collapsed string.
#'
#' @param expr Expression to evaluate.
#'
#' @noRd
#' @keywords internal
render_html <- function(expr) {
  tags <- eval(parse(text = expr))

  tmp <- tempfile(fileext = ".html")
  on.exit(expr = unlink(tmp))

  htmltools::save_html(tags, file = tmp, background = "none")

  paste0(read_lines(tmp), collapse = "")
}

#' Create a Renderer
#'
#' Create a custom renderer.
#'
#' @param fn A function that accepts two arguments,
#' the full path to the `file` to render, and the
#' `data` to render.
#'
#' @return A renderer function.
#' @examples
#' if (interactive()) {
#'   fn <- function(path, data) {
#'     # ...
#'   }
#'
#'   as_renderer(fn)
#' }
#' @export
as_renderer <- function(fn) {
  assert_that(is_function(fn))
  assert_that(is_renderer(fn))

  structure(
    fn,
    class = c(
      "renderer",
      class(fn)
    )
  )
}

#' @export
print.renderer <- function(x, ...) {
  cli::cli_alert_info("A renderer")
}

#' Is Renderer
#'
#' Check whether an object is a renderer.
#'
#' @param obj Object to check.
#'
#' @return Boolean
#' @keywords internal
#' @noRd
is_renderer_obj <- function(obj) {
  inherits(obj, "renderer")
}

#' Routing HTTP Methods
#'
#' Register route handlers for HTTP verbs on a [`Routing`] instance.
#'
#' The routing helpers provide a fluent API for attaching handlers to HTTP
#' methods. Each helper shares the same signature and behaviour.
#'
#' ## Supported helpers
#'
#' - `get()`: Respond to HTTP `GET` requests.
#' - `post()`: Respond to HTTP `POST` requests.
#' - `put()`: Respond to HTTP `PUT` requests.
#' - `patch()`: Respond to HTTP `PATCH` requests.
#' - `delete()`: Respond to HTTP `DELETE` requests.
#' - `options()`: Respond to HTTP `OPTIONS` requests.
#' - `all()`: Respond to every method above.
#'
#' @param path String. Route to listen to; use `:` to define a parameter (e.g.
#' `"/hello/:name"`).
#' @param handler Function that accepts the request and response objects and
#' returns an httpuv response (e.g. [response()]). Handlers can return the result
#' of helper functions such as `Response$text()`, `Response$json()`, or the
#' output of any renderer.
#' @param error Optional handler invoked if the route raises an error; receives
#' the request, response, and the error condition.
#'
#' @return The routing object invisibly so calls can be chained.
#'
#' @examples
#' app <- Ambiorix$new()
#'
#' app$get("/", function(req, res) {
#'   res$text("Hello, world!")
#' })
#'
#' app$post("/echo", function(req, res) {
#'   res$json(list(received = req$body))
#' })
#'
#' app$all("/health", function(req, res) {
#'   res$json(list(status = "ok"))
#' })
#'
#' @seealso [`Routing`]
#'
#' @name routing-http-methods
NULL

#' Core Routing Class
#'
#' Core routing class.
#' Do not use directly, see [Ambiorix], and [Router].
#'
#' @field error Error handler.
#' @field get Register a route handler for HTTP GET requests. See
#'   [routing-http-methods].
#' @field put Register a route handler for HTTP PUT requests. See
#'   [routing-http-methods].
#' @field patch Register a route handler for HTTP PATCH requests. See
#'   [routing-http-methods].
#' @field delete Register a route handler for HTTP DELETE requests. See
#'   [routing-http-methods].
#' @field post Register a route handler for HTTP POST requests. See
#'   [routing-http-methods].
#' @field options Register a route handler for HTTP OPTIONS requests. See
#'   [routing-http-methods].
#' @field all Register a route handler that responds to every HTTP verb used by
#'   Ambiorix. See [routing-http-methods].
#' @field basepath Basepath, read-only.
#' @field websocket Websocket handler.
#' @section HTTP methods:
#' See [routing-http-methods] for the full argument reference. The routing
#' instance exposes helpers for common HTTP verbs; they are registered when the
#' object is initialised and share the same signature.
#'
#' - `get()`, `put()`, `patch()`, `delete()`, `post()`, `options()` register a
#'   handler for the single corresponding HTTP verb; see
#'   [routing-http-methods].
#' - `all()` registers a handler that responds to `GET`, `POST`, `PUT`,
#'   `DELETE`, and `PATCH`; see [routing-http-methods].
#'
#' @return A Routing object.
#' @seealso [routing-http-methods]
#' @keywords export
Routing <- R6::R6Class(
  "Routing",
  public = list(
    error = NULL,
    get = NULL,
    put = NULL,
    patch = NULL,
    delete = NULL,
    post = NULL,
    options = NULL,
    all = NULL,
    #' @details Initialise
    #' @param path Prefix path.
    initialize = function(path = "") {
      private$.basepath <- path
      private$.is_router <- path != ""
      private$.register_http_methods()
    },
    #' @details PARAM Method
    #'
    #' @param name Name of the parameter
    #' @param handler Function that accepts the request, response, parameter value and the parameter name.
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$get("/", function(req,res){
    #'  res$send("Hello!")
    #' })
    #'
    #' app$param("person", function(req, res, value, name){
    #'  if(value == "notWanted"){
    #'   res$status <- 403L
    #'   res$send("This is the end.")
    #'  }
    #'
    #'  # continue processing the request...
    #' })
    #'
    #' app$get("/hi/:person", function(req,res){
    #'  res$sendf("Hi! %s", req$params$person)
    #' })
    #' app$get("/info/:person", function(req,res){
    #'  res$sendf("Here is all your info, %s", req$params$person)
    #' })
    #' if(interactive())
    #'  app$start()
    param = function(name, handler) {
      assert_that(not_missing(handler))
      assert_that(is_param_handler(handler))
      p <- lapply(name, function(s) {
        list(
          handler = handler,
          params = s
        )
      })
      private$.params <- append(private$.params, p)
      invisible(self)
    },
    #' @details Receive Websocket Message
    #' @param name Name of message.
    #' @param handler Function to run when message is received.
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$get("/", function(req, res){
    #'  res$send("Using {ambiorix}!")
    #' })
    #'
    #' app$receive("hello", function(msg, ws){
    #'  print(msg) # print msg received
    #'
    #'  # send a message back
    #'  ws$send("hello", "Hello back! (sent from R)")
    #' })
    #'
    #' if(interactive())
    #'  app$start()
    receive = function(name, handler) {
      private$.receivers <- append(
        private$.receivers,
        list(WebsocketHandler$new(name, handler))
      )

      invisible(self)
    },
    #' @details Print
    print = function() {
      cli::cli_rule("Ambiorix", right = "web server")
      cli::cli_li("routes: {.val {private$n_routes()}}")
    },
    #' @details Engine to use for rendering templates.
    #' @param engine Engine function.
    engine = function(engine) {
      if (!is_renderer_obj(engine)) {
        engine <- as_renderer(engine)
      }

      self$use(engine)
      invisible(self)
    },
    #' @details Use a router or middleware
    #' @param use Either a router as returned by [Router], a function to use as middleware,
    #' or a `list` of functions.
    #' If a function is passed, it must accept two arguments (the request, and the response):
    #' this function will be executed every time the server receives a request.
    #' _Middleware may but does not have to return a response, unlike other methods such as `get`_
    #' Note that multiple routers and middlewares can be used.
    use = function(use) {
      assert_that(not_missing(use))

      # recurse through items
      if (is.list(use)) {
        for (i in seq_along(use)) {
          self$use(use[[i]])
        }
      }

      # mount router
      if (inherits(use, "Router")) {
        private$.routers <- append(private$.routers, use)
      }

      if (is_renderer_obj(use) && private$.is_router) {
        .globals$errorLog$log(
          "Cannot pass cookie preprocessor to `Router`, only to `Ambiorix`"
        )
        return(invisible(self))
      }

      if (is_renderer_obj(use)) {
        .Deprecated(
          "engine",
          package = "ambiorix",
          msg = "Use `engine` instead of `use` for custom renderers."
        )
        .globals$renderer <- use
        return(invisible(self))
      }

      if (is_cookie_parser(use) && private$.is_router) {
        .globals$errorLog$log(
          "Cannot pass cookie parser to `Router`, only to `Ambiorix`"
        )
        return(invisible(self))
      }

      if (is_path_to_pattern(use) && private$.is_router) {
        .globals$errorLog$log(
          "Cannot pass path to pattern converter to `Router`, only to `Ambiorix`"
        )
        return(invisible(self))
      }

      if (is_cookie_parser(use)) {
        .globals$cookieParser <- use
        return(invisible(self))
      }

      if (is_path_to_pattern(use)) {
        .globals$pathToPattern <- use
        return(invisible(self))
      }

      if (is_cookie_preprocessor(use) && private$.is_router) {
        .globals$errorLog$log(
          "Cannot pass cookie preprocessor to `Router`, only to `Ambiorix`"
        )
        return(invisible(self))
      }

      if (is_cookie_preprocessor(use)) {
        .globals$cookiePreprocessors <- append(
          .globals$cookiePreprocessors,
          use
        )
        return(invisible(self))
      }

      # pass middleware
      if (is.function(use)) {
        assert_that(is_handler(use))
        private$.middleware <- append(private$.middleware, use)
        return(invisible(self))
      }

      invisible(self)
    },
    #' @details Get the routes
    #' @param routes Existing list of routes.
    #' @param parent Parent path.
    get_routes = function(routes = list(), parent = "") {
      routes <- append(
        routes,
        lapply(
          private$.routes,
          function(route) {
            route$route$as_pattern(parent)
            route$route$decompose(parent)
            route$route$basepath <- private$.basepath
            route
          }
        )
      )

      if (!length(private$.routers)) {
        return(routes)
      }

      parent <- paste0(parent, private$.basepath)

      for (router in private$.routers) {
        routes <- router$get_routes(routes, parent)
      }

      return(routes)
    },
    #' @details Get the parameter middlewares
    #' @param params Existing list of parameter middlewares.
    #' @param parent Parent path.
    get_params = function(params = list(), parent = "") {
      params <- append(
        params,
        lapply(
          private$.params,
          function(fn) {
            attr(fn, "basepath") <- private$.basepath
            return(fn)
          }
        )
      )

      if (!length(private$.routers)) {
        return(params)
      }

      parent <- paste0(parent, private$.basepath)

      for (router in private$.routers) {
        params <- router$get_params(params, parent)
      }

      return(params)
    },
    #' @details Get the websocket receivers
    #' @param receivers Existing list of receivers
    get_receivers = function(receivers = list()) {
      receivers <- append(receivers, private$.receivers)

      if (!length(private$.routers)) {
        return(receivers)
      }

      for (router in private$.routers) {
        receivers <- router$get_receivers(receivers)
      }

      return(receivers)
    },
    #' @details Get the middleware
    #' @param middlewares Existing list of middleswares
    #' @param parent Parent path
    get_middleware = function(middlewares = list(), parent = "") {
      middlewares <- append(
        middlewares,
        lapply(
          private$.middleware,
          function(fn) {
            attr(fn, "basepath") <- paste0(parent, private$.basepath)
            return(fn)
          }
        )
      )

      if (!length(private$.routers)) {
        return(middlewares)
      }

      parent <- paste0(parent, private$.basepath)

      for (router in private$.routers) {
        middlewares <- router$get_middleware(middlewares, parent)
      }

      return(middlewares)
    },
    #' @details Prepare routes and decomposes paths
    prepare = function() {
      for (route in private$.routes) {
        route$route$as_pattern()
        route$route$decompose()
      }

      private$reorder_routes()
      if (!length(private$.routers)) {
        return()
      }

      for (route in private$.routers) {
        route$prepare()
      }
    }
  ),
  active = list(
    basepath = function(path) {
      if (!missing(path)) {
        private$.basepath <- path
        return(path)
      }

      invisible(private$.basepath)
    },
    websocket = function(ws) {
      if (missing(ws) && !is.null(private$.wss_custom)) {
        return(private$.wss_custom)
      }

      if (missing(ws) && is.null(private$.wss_custom)) {
        return(private$.wss)
      }

      private$.wss_custom <- ws
      invisible(self)
    }
  ),
  private = list(
    .basepath = "/",
    .is_router = FALSE,
    .routes = list(),
    .static = list(),
    .receivers = list(),
    .middleware = list(),
    .params = list(),
    .is_running = FALSE,
    .wss_custom = NULL,
    .routers = list(),
    .register_http_methods = function() {
      http_methods = list(
        get = "GET",
        put = "PUT",
        patch = "PATCH",
        delete = "DELETE",
        post = "POST",
        options = "OPTIONS",
        all = c("GET", "POST", "PUT", "DELETE", "PATCH")
      )

      Map(
        f = function(name, value) {
          self[[name]] <- function(path, handler, error = NULL) {
            assert_that(valid_path(path))
            assert_that(not_missing(handler))
            assert_that(is_handler(handler))

            r <- list(
              route = Route$new(private$.make_path(path)),
              path = path,
              fun = handler,
              method = value,
              error = error %error% self$error
            )
            private$.routes <- append(private$.routes, list(r))

            invisible(self)
          }
        },
        name = names(http_methods),
        value = unname(http_methods)
      )

      invisible(self)
    },
    # we reorder the routes before launching the app
    # we make sure the longest patterns are checked first
    # this makes sure /:id/x matches BEFORE /:id does
    # however we also want to try to match exact paths
    # BEFORE dynamic ones
    # e.g. /hello should be matched before /:id
    reorder_routes = function() {
      if (!length(private$.routes)) {
        return()
      }

      indices <- seq_along(private$.routes)
      paths <- lapply(private$.routes, function(route) {
        data.frame(
          nchar = nchar(route$route$path),
          dynamic = route$route$dynamic
        )
      })
      df <- do.call(rbind, paths)
      df$order <- seq_len(nrow(df))
      df <- df[order(df$dynamic, -df$nchar), ]

      private$.routes <- private$.routes[df$order]
    },
    .call = function(req) {
      request <- Request$new(req)
      res <- Response$new()

      # loop over routes
      for (i in seq_along(private$.routes)) {
        # if path matches pattern and method
        if (
          grepl(private$.routes[[i]]$route$pattern, req$PATH_INFO) &&
            req$REQUEST_METHOD %in% private$.routes[[i]]$method
        ) {
          .globals$infoLog$log(req$REQUEST_METHOD, "on", req$PATH_INFO)

          basepath <- private$.routes[[i]]$route$basepath

          # parse request
          request$params <- tryCatch(
            set_params(request$PATH_INFO, private$.routes[[i]]$route),
            error = function(error) {
              error
            }
          )

          if (inherits(request$params, "error")) {
            return(private$.routes[[i]]$error(request, res, request$params))
          }

          # parameter middleware

          if (length(private$.params) > 0L && length(request$params) > 0L) {
            for (j in seq_along(private$.params)) {
              param_res <- NULL
              pn <- private$.params[[j]]$params
              pv <- request$params[[pn]]

              # if param middleware is on correct router and has a handler for
              # a request parameter.

              if (
                identical(attr(private$.params[[j]], "basepath"), basepath) &&
                  !is.null(pv)
              ) {
                param_res <- private$.params[[j]]$handler(
                  request,
                  res,
                  pv,
                  pn
                )

                if (is_response(param_res)) {
                  return(param_res)
                }
              }
            }
          }

          # Middleware

          if (length(private$.middleware) > 0L) {
            for (j in seq_along(private$.middleware)) {
              mid_basepath <- attr(private$.middleware[[j]], "basepath")

              mid_res <- NULL
              if (grepl(mid_basepath, req$PATH_INFO)) {
                mid_res <- private$.middleware[[j]](request, res)
              }

              if (is_response(mid_res)) {
                return(mid_res)
              }
            }
          }

          # get response
          response <- tryCatch(
            private$.routes[[i]]$fun(request, res),
            error = function(error) {
              error
            }
          )

          if (
            inherits(response, "error") && !is.null(private$.routes[[i]]$error)
          ) {
            return(private$.routes[[i]]$error(request, res, response))
          }

          if (inherits(response, "error") && !is.null(self$error)) {
            return(self$error(request, res, response))
          }

          if (inherits(x = response, what = c("promise", "Future", "mirai"))) {
            return(
              promises::then(
                response,
                onFulfilled = function(response) {
                  response %response%
                    response("Must return a response", status = 206L)
                },
                onRejected = function(error) {
                  message(conditionMessage(error))
                  .globals$errorLog$log(
                    req$REQUEST_METHOD,
                    "on",
                    req$PATH_INFO,
                    "-",
                    "Server error"
                  )
                  private$.routes[[i]]$error(request, res, error)
                }
              )
            )
          }

          if (is_forward(response)) {
            next
          }

          # if not a response return something that is
          return(
            response %response%
              response("Must return a response", status = 206L)
          )
        }
      }

      .globals$errorLog$log(
        request$REQUEST_METHOD,
        "on",
        request$PATH_INFO,
        "- Not found"
      )

      # return 404
      request$params <- set_params(
        request$PATH_INFO,
        Route$new(request$PATH_INFO)
      )

      self$not_found(request, res)
    },
    .wss = function(ws) {
      .globals$wsc <- append(.globals$wsc, Websocket$new(ws))

      # receive
      ws$onMessage(function(binary, message) {
        # don't run if no receiver
        if (length(private$.receivers) == 0) {
          return(NULL)
        }

        message <- yyjsonr::read_json_str(message)

        for (i in seq_along(private$.receivers)) {
          if (private$.receivers[[i]]$is_handler(message)) {
            .globals$infoLog$log(
              "Received message from websocket:",
              message$name
            )
            return(private$.receivers[[i]]$receive(message, ws))
          }
        }
      })
    },
    n_routes = function() {
      length(private$.routes)
    },
    .make_path = function(path) {
      paste0(private$.basepath, path)
    }
  )
)

test_that("application", {
  library(ambiorix)

  app <- Ambiorix$new()

  app$get("/", function(req, res) {
    res$send("home")
  })

  app$post("/", function(req, res) {
    res$send("home")
  })

  app$put("/", function(req, res) {
    res$send("home")
  })

  app$patch("/", function(req, res) {
    res$send("home")
  })

  app$delete("/", function(req, res) {
    res$send("home")
  })

  app$all("/", function(req, res) {
    res$send("home")
  })

  app$options("/", function(req, res) {
    res$send("home")
  })

  # dynamic
  app$get("/.path", function(req, res) {
    res$send("home")
  })

  app$receive("message", function(...) {
    print("received")
  })

  expect_type(app$get_routes(), "list")
  expect_length(app$get_routes(), 8L)
  expect_snapshot(app)

  stop_all()
})

test_that("Websocket", {
  expect_error(WebsocketHandler$new())
  ws <- WebsocketHandler$new("hello", function(msg) {
    return(msg)
  })
  msg <- ws$receive(list(message = "world"))
  expect_equal(msg, "world")
  expect_snapshot(ws)

  ws <- Websocket$new(list())
  expect_snapshot(ws)
})

test_that("parameter middleware", {
  app <- Ambiorix$new()

  app$param("uid", function(req, res, value, name) {
    req$params$uid <- "loki"
  })
  # two parameters with the same callback
  app$param(c("from", "to"), function(req, res, value, name) {
    req$params[[name]] <- as.numeric(value)
  })

  app$get("/user/:uid", function(req, res) {
    res$send(req$params$uid)
  })
  app$get("/post/:from/:to", function(req, res) {
    from <- typeof(req$params$from)
    to <- typeof(req$params$to)
    res$send(paste0(from, "-", to))
  })

  expect_type(app$get_params(), "list")
  expect_length(app$get_params(), 3L)

  stop_all()
})

.globals <- new.env(hash = TRUE)
.cache_tmpls <- new.env(hash = TRUE)

.onLoad <- function(libname, pkgname) {
  .globals$infoLog <- new_log(info())
  .globals$errorLog <- new_log(error())
  .globals$successLog <- new_log(success())
  .globals$cookieParser <- default_cookie_parser
  .globals$pathToPattern <- NULL
  .globals$cookiePreprocessors <- list()
  .globals$renderer <- NULL
  .globals$cache_tmpls <- FALSE
  .globals$wsc <- list()
}

#' Serialise
#'
#' Serialise data to JSON.
#'
#' @param data Data to serialise.
#' @param ... Named options to pass to [yyjsonr::write_json_str].
#'
#' @noRd
#' @keywords internal
default_serialiser <- function(data, ...) {
  dots <- list(...)

  # `yyjsonr::write_json_str()` accepts both `opts` & `...` but
  # `...` should override `opts`.
  # ensure that happens and use `opts` only:
  opts <- dots$opts
  if (is.null(opts)) {
    opts <- list()
  }

  dots$opts <- NULL
  opts[names(dots)] <- dots

  if (is.null(opts$auto_unbox)) {
    opts$auto_unbox <- TRUE
  }

  yyjsonr::write_json_str(data, opts = opts)
}

#' Retrieve Serialiser
#'
#' Retrieve the serialiser to use, either the default or that defined by user.
#'
#' @noRd
#' @keywords internal
get_serialise <- function() {
  getOption("AMBIORIX_SERIALISER", default_serialiser)
}

#' Serialise an Object to JSON
#'
#' @details
#' Ambiorix uses [yyjsonr::write_json_str()] by default for serialization.
#'
#' ### Custom Serialiser
#'
#' To override the default, set the `AMBIORIX_SERIALISER` option to a function that accepts:
#' - `data`: Object to serialise.
#' - `...`: Additional arguments passed to the function.
#'
#' For example:
#'
#' ```r
#' my_serialiser <- function(data, ...) {
#'  jsonlite::toJSON(x = data, ...)
#' }
#'
#' options(AMBIORIX_SERIALISER = my_serialiser)
#' ```
#'
#' @param data Data to serialise.
#' @param ... Passed to serialiser.
#'
#' @examples
#' if (interactive()) {
#'   # a list:
#'   response <- list(code = 200L, msg = "hello, world!")
#'
#'   serialise(response)
#'   #> {"code":200,"msg":"hello, world"}
#'
#'   serialise(response, auto_unbox = FALSE)
#'   #> {"code":[200],"msg":["hello, world"]}
#'
#'   # data.frame:
#'   serialise(cars)
#' }
#'
#' @return JSON string.
#' @export
serialise <- function(data, ...) {
  get_serialise()(data, ...)
}

test_that("serialise", {
  obj <- list(a = 1, b = "hello, world!")

  expect_equal(
    default_serialiser(obj),
    yyjsonr::write_json_str(obj, auto_unbox = TRUE)
  )

  expect_equal(
    default_serialiser(obj, auto_unbox = FALSE),
    yyjsonr::write_json_str(obj)
  )

  expect_equal(
    default_serialiser(obj, opts = list(auto_unbox = FALSE)),
    yyjsonr::write_json_str(obj)
  )
})

test_that("serialise falls back to default serialiser when option unset", {
  obj <- list(a = 1, b = "hello, world!")
  global_serialiser <- getOption("AMBIORIX_SERIALISER")
  on.exit(options(AMBIORIX_SERIALISER = global_serialiser), add = TRUE)

  options(AMBIORIX_SERIALISER = NULL)

  expect_identical(
    serialise(obj),
    default_serialiser(obj)
  )
})

test_that("serialise forwards dots to custom serialiser", {
  obj <- list(a = 1)
  global_serialiser <- getOption("AMBIORIX_SERIALISER")
  on.exit(options(AMBIORIX_SERIALISER = global_serialiser), add = TRUE)

  captured <- NULL
  my_serialiser <- function(data, ...) {
    captured <<- list(data = data, dots = list(...))
    "custom"
  }

  options(AMBIORIX_SERIALISER = my_serialiser)

  result <- serialise(obj, digits = 2, auto_unbox = FALSE)
  expect_equal(result, "custom")
  expect_equal(captured$data, obj)
  expect_equal(captured$dots, list(digits = 2, auto_unbox = FALSE))
})

test_that("default serialiser gives precedence to dots over opts", {
  obj <- list(a = 1)

  expect_equal(
    default_serialiser(
      obj,
      opts = list(auto_unbox = FALSE),
      auto_unbox = TRUE
    ),
    yyjsonr::write_json_str(
      obj,
      opts = list(auto_unbox = TRUE)
    )
  )
})

test_that("custom serialiser works", {
  global_serialiser <- getOption("AMBIORIX_SERIALISER")
  on.exit(options(AMBIORIX_SERIALISER = global_serialiser))

  my_serialiser <- function(data, ...) {
    list(
      data = data,
      serialised = jsonlite::toJSON(x = data, ...)
    )
  }

  options(AMBIORIX_SERIALISER = my_serialiser)

  obj <- list(a = 1, b = "hello, world!")

  expect_equal(
    my_serialiser(obj),
    list(
      data = obj,
      serialised = jsonlite::toJSON(obj)
    )
  )
})

#' Ambiorix
#'
#' Web server.
#'
#' @field not_found 404 Response, must be a handler function that accepts the request and the response, by default uses [response_404()].
#' @field error 500 response when the route errors, must a handler function that accepts the request and the response, by default uses [response_500()].
#' @field on_stop Callback function to run when the app stops, takes no argument.
#' @field port Port to run the application.
#' @field host Host to run the application.
#' @field limit Max body size, defaults to `5 * 1024 * 1024`.
#'
#' @importFrom assertthat assert_that
#' @importFrom utils browseURL
#' @importFrom methods formalArgs
#'
#' @examples
#' app <- Ambiorix$new()
#'
#' app$get("/", function(req, res){
#'  res$send("Using {ambiorix}!")
#' })
#'
#' app$on_stop <- function(){
#'  cat("Bye!\n")
#' }
#'
#' if(interactive())
#'  app$start()
#'
#' @return An object of class `Ambiorix` from which one can
#' add routes, routers, and run the application.
#'
#' @export
Ambiorix <- R6::R6Class(
  "Ambiorix",
  inherit = Routing,
  public = list(
    not_found = NULL,
    error = NULL,
    on_stop = NULL,
    #' @details Define the webserver.
    #'
    #' @param host A string defining the host.
    #' @param port Integer defining the port, defaults to `ambiorix.port` option: uses a random port if `NULL`.
    #' @param log Whether to generate a log of events.
    initialize = function(
      host = getOption("ambiorix.host", "0.0.0.0"),
      port = getOption("ambiorix.port", NULL),
      log = getOption("ambiorix.logger", TRUE)
    ) {
      super$initialize()
      .globals$infoLog$predicate <- logPredicate(log)
      .globals$errorLog$predicate <- logPredicate(log)
      .globals$successLog$predicate <- logPredicate(log)

      private$.host <- host
      private$.port <- get_port(host, port)

      self$not_found <- function(req, res) {
        response_404()
      }

      self$error <- function(req, res, error) {
        message(conditionMessage(error))
        res$status <- 500L
        res$send("500: Internal Server Error")
      }

      invisible(self)
    },
    #' @details Cache templates in memory instead of reading
    #' them from disk.
    cache_templates = function() {
      .globals$cache_tmpls <- TRUE
      invisible(self)
    },
    #' @details Specifies the port to listen on.
    #' @param port Port number.
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$listen(3000L)
    #'
    #' app$get("/", function(req, res){
    #'  res$send("Using {ambiorix}!")
    #' })
    #'
    #' if(interactive())
    #'  app$start()
    listen = function(port) {
      assert_that(not_missing(port))
      private$.port <- as.integer(port)
      invisible(self)
    },
    #' @details Sets the 404 page.
    #' @param handler Function that accepts the request and returns an object
    #' describing an httpuv response, e.g.: [response()].
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$set_404(function(req, res){
    #'  res$send("Nothing found here")
    #' })
    #'
    #' app$get("/", function(req, res){
    #'  res$send("Using {ambiorix}!")
    #' })
    #'
    #' if(interactive())
    #'  app$start()
    set_404 = function(handler) {
      assert_that(not_missing(handler))
      assert_that(is_handler(handler))
      self$not_found <- handler
      invisible(self)
    },
    #' @details Sets the error handler.
    #' @param handler Function that accepts a request, response and an error object.
    #'
    #' @examples
    #' # my custom error handler:
    #' error_handler <- function(req, res, error) {
    #'   if (!is.null(error)) {
    #'     error_msg <- conditionMessage(error)
    #'     cli::cli_alert_danger("Error: {error_msg}")
    #'   }
    #'   response <- list(
    #'     code = 500L,
    #'     msg = "Uhhmmm... Looks like there's an error from our side :("
    #'   )
    #'   res$
    #'     set_status(500L)$
    #'     json(response)
    #' }
    #'
    #' # handler for GET at /whoami:
    #' whoami <- function(req, res) {
    #'   # simulate error (object 'Pikachu' is not defined)
    #'   print(Pikachu)
    #' }
    #'
    #' app <- Ambiorix$
    #'   new()$
    #'   set_error(error_handler)$
    #'   get("/whoami", whoami)
    #'
    #' if (interactive()) {
    #'   app$start(open = FALSE)
    #' }
    set_error = function(handler) {
      assert_that(not_missing(handler))
      assert_that(is_error_handler(handler))
      self$error <- handler
      invisible(self)
    },
    #' @details Static directories
    #'
    #' @param path Local path to directory of assets.
    #' @param uri URL path where the directory will be available.
    static = function(path, uri = "www") {
      assert_that(not_missing(uri))
      assert_that(not_missing(path))

      lst <- list(path)
      names(lst) <- uri
      private$.static <- append(private$.static, lst)
      invisible(self)
    },
    #' @details Start
    #' Start the webserver.
    #' @param host A string defining the host.
    #' @param port Integer defining the port, defaults to `ambiorix.port` option: uses a random port if `NULL`.
    #' @param open Whether to open the app the browser.
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$get("/", function(req, res){
    #'  res$send("Using {ambiorix}!")
    #' })
    #'
    #' if(interactive())
    #'  app$start(port = 3000L)
    start = function(
      port = NULL,
      host = NULL,
      open = interactive()
    ) {
      if (private$.is_running) {
        cli::cli_alert_warning("Server is already running")
        return()
      }
      if (is.null(port)) {
        port <- private$.port
      }

      if (is.null(host)) {
        host <- private$.host
      }

      port <- get_port(host, port)

      super$prepare()
      private$.routes <- super$get_routes()

      if (private$n_routes() == 0L) {
        stop("No routes specified")
      }

      private$.receivers <- super$get_receivers()
      private$.middleware <- super$get_middleware()
      private$.params <- super$get_params()

      private$.server <- httpuv::startServer(
        host = host,
        port = port,
        app = list(
          call = super$.call,
          staticPaths = private$.static,
          onWSOpen = super$websocket,
          staticPathOptions = httpuv::staticPathOptions(
            html_charset = "utf-8",
            headers = list(
              "X-UA-Compatible" = "IE=edge,chrome=1"
            )
          ),
          onHeaders = function(req) {
            size <- 0L
            if (private$.limit <= 0) {
              return(NULL)
            }

            if (length(req$CONTENT_LENGTH) > 0) {
              size <- as.numeric(req$CONTENT_LENGTH)
            } else if (length(req$HTTP_TRANSFER_ENCODING) > 0) {
              size <- Inf
            }

            if (size > private$.limit) {
              .globals$errorLog$log("Request size exceeded, see app$limit")

              return(
                response(
                  "Maximum upload size exceeded",
                  status = 413L,
                  headers = list("Content-Type" = "text/plain")
                )
              )
            }

            return(NULL)
          }
        )
      )

      browser_host <- switch(
        EXPR = host,
        "0.0.0.0" = "127.0.0.1",
        host
      )

      browser_url <- sprintf("http://%s:%s", browser_host, port)

      .globals$successLog$log("Listening on", browser_url)

      # runs
      private$.is_running <- TRUE

      # open
      browse_ambiorix(open, browser_url)

      on.exit({
        self$stop()
      })

      # continually process requests:
      httpuv::service(timeoutMs = Inf)

      invisible(self)
    },
    #' @details Define Serialiser
    #' @param handler Function to use to serialise.
    #' This function should accept two arguments: the object to serialise and `...`.
    #'
    #' @examples
    #' app <- Ambiorix$new()
    #'
    #' app$serialiser(function(data, ...){
    #'  jsonlite::toJSON(x, ..., pretty = TRUE)
    #' })
    #'
    #' app$get("/", function(req, res){
    #'  res$send("Using {ambiorix}!")
    #' })
    #'
    #' if(interactive())
    #'  app$start()
    serialiser = function(handler) {
      assert_that(is_function(handler))
      options(AMBIORIX_SERIALISER = handler)
      invisible(self)
    },
    #' @details Stop
    #' Stop the webserver.
    stop = function() {
      if (!private$.is_running) {
        .globals$errorLog$log("Server not running")
        return(invisible())
      }

      # run on stop
      if (!is.null(self$on_stop)) {
        self$on_stop()
      }

      private$.server$stop()
      .globals$errorLog$log("Server stopped")

      private$.is_running <- FALSE

      invisible(self)
    },
    #' @details Print
    print = function() {
      cli::cli_rule("Ambiorix", right = "web server")
      cli::cli_li("routes: {.val {private$n_routes()}}")
    }
  ),
  active = list(
    port = function(value) {
      if (missing(value)) {
        return(private$.port)
      }

      private$.port <- as.integer(value)
    },
    host = function(value) {
      if (missing(value)) {
        return(private$.host)
      }

      private$.host <- value
    },
    limit = function(value) {
      if (missing(value)) {
        return(private$.limit)
      }

      private$.limit <- as.integer(value)
    }
  ),
  private = list(
    .host = "0.0.0.0",
    .port = 3000,
    .server = NULL,
    .static = list(),
    .is_running = FALSE,
    .limit = 5 * 1024 * 1024,
    n_routes = function() {
      length(private$.routes) + length(private$.static)
    },
    .make_path = function(path) {
      paste0(private$.basepath, path)
    }
  )
)

test_that("Ambiorix", {
  # default
  app <- Ambiorix$new()
  expect_s3_class(app, "Ambiorix")

  # bindings
  app <- Ambiorix$new()
  app$port <- 3000L
  app$host <- "xxx"
  expect_equal(app$port, 3000L)
  expect_equal(app$host, "xxx")

  # set
  app <- Ambiorix$new(
    port = 8080L,
    host = "127.0.0.1",
    log = FALSE
  )

  app$static("www")
  expect_equal(app$port, 8080L)
  expect_equal(app$host, "127.0.0.1")

  expect_error(app$set_404("error"))
  expect_error(app$set_404(function(req) {}))
  expect_s3_class(
    app$set_404(function(req, res) {
      res$send("Errr")
    }),
    "Ambiorix"
  )

  expect_error(app$serialiser("error"))
  expect_s3_class(
    app$serialiser(function(data) {
      return("data")
    }),
    "Ambiorix"
  )
})

#' Import Files
#'
#' Import all R-files in a directory.
#'
#' @param ... Directory from which to import `.R` or `.r` files.
#'
#' @examples
#' if (interactive()) {
#'   import("views")
#' }
#'
#' @return Invisibly returns `NULL`.
#'
#' @export
import <- function(...) {
  files <- list.files(..., pattern = "\\.R$|\\.r$", full.names = TRUE)
  sapply(files, source)
  invisible()
}

valid_path <- function(x) {
  if (missing(x)) {
    return(FALSE)
  }

  if (!inherits(x, "character")) {
    return(FALSE)
  }

  return(TRUE)
}

assertthat::on_failure(valid_path) <- function(call, env) {
  paste0(deparse(call$x), " is not valid")
}

not_missing <- function(x) {
  !missing(x)
}

assertthat::on_failure(not_missing) <- function(call, env) {
  paste("Missing", deparse(call$x))
}

has_file <- function(x) {
  file.exists(x)
}

assertthat::on_failure(has_file) <- function(call, env) {
  paste("Cannot find", deparse(call$x))
}

is_handler <- function(x) {
  is_fun <- is.function(x)
  has_args <- length(formalArgs(x)) == 2

  all(is_fun, has_args)
}

assertthat::on_failure(is_handler) <- function(call, env) {
  paste("`handler` must be a function that accepts: `req`, and `res`")
}

is_param_handler <- function(x) {
  is_fun <- is.function(x)
  has_args <- length(formalArgs(x)) == 4
  all(is_fun, has_args)
}

assertthat::on_failure(is_param_handler) <- function(call, env) {
  paste(
    "`handler` must be a function that accepts: `req`, `res`, `value` and `name`"
  )
}

is_error_handler <- function(x) {
  is_fun <- is.function(x)
  has_args <- length(formalArgs(x)) == 3

  all(is_fun, has_args)
}

assertthat::on_failure(is_error_handler) <- function(call, env) {
  "`handler` must be a function that accepts: `req`, `res` and `error`"
}

is_logger <- function(x) {
  inherits(x, "Logger")
}

assertthat::on_failure(is_logger) <- function(call, env) {
  sprintf(
    "%s must be an object of class `Logger` from the `log` package",
    deparse(call$x)
  )
}

is_function <- function(x) {
  is.function(x)
}

assertthat::on_failure(is_function) <- function(call, env) {
  sprintf(
    "`%s` is not a function",
    deparse(call$x)
  )
}

is_renderer <- function(x) {
  length(formalArgs) != 2L
}

assertthat::on_failure(is_renderer) <- function(call, env) {
  sprintf(
    "`%s` must accept two arguments: `file` and `data`",
    deparse(call$x)
  )
}

#' Cookie Parser
#'
#' Parses the cookie string.
#'
#' @param req A [Request].
#' @examples
#' if (interactive()) {
#'   library(ambiorix)
#'
#'   #' Handle GET at '/greet'
#'   #'
#'   #' @export
#'   say_hello <- function(req, res) {
#'     cookies <- default_cookie_parser(req)
#'     print(cookies)
#'
#'     res$send("hello there!")
#'   }
#'
#'   app <- Ambiorix$new()
#'   app$get("/greet", say_hello)
#'   app$start()
#' }
#'
#' @return A `list` of key value pairs or cookie values.
#' @export
default_cookie_parser <- function(req) {
  cookie_new <- list()

  if (is.null(req$HTTP_COOKIE)) {
    return(cookie_new)
  }

  if (req$HTTP_COOKIE == "") {
    return(list())
  }

  split <- strsplit(req$HTTP_COOKIE, ";")[[1]]
  split <- strsplit(split, "=")
  for (i in seq_along(split)) {
    value <- trimws(split[[i]])

    if (length(value) < 2) {
      next
    }

    if (value[1] == "") {
      next
    }

    cookie_new[[value[1]]] <- value[2]
  }

  return(cookie_new)
}

#' Define a Cookie Parser
#'
#' Identifies a function as a cookie parser (see example).
#'
#' @param fn A function that accepts a single argument,
#' `req` the [Request] and returns the parsed cookie string,
#' generally a `list`.
#' Note that the original cookie string is available on the
#' [Request] at the `HTTP_COOKIE` field, get it with:
#' `req$HTTP_COOKIE`
#'
#' @examples
#' func <- function(req) {
#'  req$HTTP_COOKIE
#' }
#'
#' parser <- as_cookie_parser(func)
#'
#' app <- Ambiorix$new()
#' app$use(parser)
#'
#' @return Object of class "cookieParser".
#' @export
as_cookie_parser <- function(fn) {
  assert_that(not_missing(fn))
  assert_that(is_function(fn))

  fn <- structure(
    fn,
    class = c(
      "cookieParser",
      class(fn)
    )
  )

  invisible(fn)
}

#' @export
print.cookieParser <- function(x, ...) {
  cli::cli_alert_info("A cookie parser")
}

#' @keywords internal
#' @noRd
is_cookie_parser <- function(obj) {
  inherits(obj, "cookieParser")
}

#' Define a Cookie Preprocessor
#'
#' Identifies a function as a cookie preprocessor.
#'
#' @param fn A function that accepts the same arguments
#' as the `cookie` method of the [Response] class
#' (name, value, ...), and returns a modified
#' `value`.
#'
#' @examples
#' func <- function(name, value, ...) {
#'  sprintf("prefix.%s", value)
#' }
#'
#' prep <- as_cookie_preprocessor(func)
#'
#' app <- Ambiorix$new()
#' app$use(prep)
#'
#' @return Object of class "cookiePreprocessor".
#' @export
as_cookie_preprocessor <- function(fn) {
  assert_that(not_missing(fn))
  assert_that(is_function(fn))

  fn <- structure(
    fn,
    class = c(
      "cookiePreprocessor",
      class(fn)
    )
  )

  invisible(fn)
}

#' @export
print.cookiePreprocessor <- function(x, ...) {
  cli::cli_alert_info("A cookie pre-processor")
}

#' @keywords internal
is_cookie_preprocessor <- function(obj) {
  inherits(obj, "cookiePreprocessor")
}

#' Cookie
#'
#' Create a cookie object.
#'
#' @param name Name of the cookie.
#' @param value value of the cookie.
#' @param expires Expiry, if an integer assumes it's the number of seconds
#' from now. Otherwise accepts an object of class `POSIXct` or `Date`.
#' If a `character` string then it is set as-is and not pre-processed.
#' If unspecified, the cookie becomes a session cookie. A session finishes
#' when the client shuts down, after which the session cookie is removed.
#' @param max_age Indicates the number of seconds until the cookie expires.
#' A zero or negative number will expire the cookie immediately.
#' If both `expires` and `max_age` are set, the latter has precedence.
#' @param domain Defines the host to which the cookie will be sent.
#' If omitted, this attribute defaults to the host of the current document URL,
#' not including subdomains.
#' @param path Indicates the path that must exist in the requested URL for the
#' browser to send the Cookie header.
#' @param secure Indicates that the cookie is sent to the server only when a
#' request is made with the https: scheme (except on localhost), and therefore,
#' is more resistant to man-in-the-middle attacks.
#' @param http_only Forbids JavaScript from accessing the cookie, for example,
#' through the document.cookie property.
#' @param same_site Controls whether or not a cookie is sent with cross-origin
#' requests, providing some protection against cross-site request forgery
#' attacks (CSRF). Accepts `Strict`, `Lax`, or `None`.
#'
#' @keywords internal
#' @noRd
cookie <- function(
  name,
  value,
  expires = NULL,
  max_age = NULL,
  domain = NULL,
  path = NULL,
  secure = TRUE,
  http_only = TRUE,
  same_site = NULL
) {
  opts <- as.list(environment())
  structure(
    opts,
    class = c(
      "cookie",
      class(opts)
    )
  )
}

#' @export
print.cookie <- function(x, ...) {
  cli::cli_alert_info("A cookie: {.field {x$name}} = {.val  {x$value}}")
}

test_that("Request cookie", {
  # cookie set
  req <- mockRequest(
    cookie = "yummy_cookie=choco; tasty_cookie=strawberry"
  )
  expect_s3_class(req, "Request")
  expect_length(req$cookie, 2L)
  expect_equal(
    req$cookie,
    list(
      yummy_cookie = "choco",
      tasty_cookie = "strawberry"
    )
  )

  # empty coookie
  req <- mockRequest(cookie = "")
  expect_length(req$cookie, 0L)
  expect_type(req$cookie, "list")
  expect_equal(req$cookie, list())

  # incorrect cookie
  req <- mockRequest(
    cookie = "yummy_cookie=choco; tasty_cookie="
  )
  expect_length(req$cookie, 1L)
  expect_type(req$cookie, "list")
  expect_equal(
    req$cookie,
    list(
      yummy_cookie = "choco"
    )
  )

  # parser
  fn <- function(req) {
    return(req$HTTP_COOKIE)
  }
  expect_false(is_cookie_parser(fn))
  parser <- as_cookie_parser(fn)
  expect_true(is_cookie_parser(parser))
  expect_snapshot(parser)
  app <- Ambiorix$new()
  app$use(parser)
  req <- mockRequest(
    cookie = "yummy_cookie=choco;"
  )
  expect_type(req$cookie, "character")
  expect_equal(req$cookie, "yummy_cookie=choco;")

  # object
  cook <- cookie("hello", "world")
  expect_snapshot(cook)
  expect_s3_class(cook, "cookie")

  # expires date
  res <- Response$new()
  res$cookie("hello", "world", expires = as.Date("2022-01-01"))
  resp <- res$send("hello")
  expect_equal(
    res$headers[["Set-Cookie"]],
    "hello=world; Expires=Sat, 01 Jan 2022 00:00:00 GMT; Path=/; Secure; HttpOnly"
  )

  # expires character
  res <- Response$new()
  res$cookie("hello", "world", expires = "2022-01-01")
  resp <- res$send("hello")
  expect_equal(
    res$headers[["Set-Cookie"]],
    "hello=world; Expires=2022-01-01; Path=/; Secure; HttpOnly"
  )

  # preprocessor
  .fn <- function(name, value, ...) {
    sprintf("prefix.%s", value)
  }
  prep <- as_cookie_preprocessor(.fn)

  app <- Ambiorix$new()
  app$use(prep)
  res <- Response$new()
  res$cookie("hello", "world")
  resp <- res$send("hello")
  expect_equal(
    res$headers[["Set-Cookie"]],
    "hello=prefix.world; Path=/; Secure; HttpOnly"
  )

  # clear
  res$cookie("world", "hello")
  res$clear_cookie("world")
  resp <- res$send("hello")
  expect_equal(
    res$headers[["Set-Cookie"]],
    "hello=prefix.world; Path=/; Secure; HttpOnly"
  )
})

#' Plain Responses
#'
#' Plain HTTP Responses.
#'
#' @param body Body of response.
#' @param headers HTTP headers.
#' @param status Response status
#'
#' @examples
#' app <- Ambiorix$new()
#'
#' # html
#' app$get("/", function(req, res){
#'  res$send("hello!")
#' })
#'
#' # text
#' app$get("/text", function(req, res){
#'  res$text("hello!")
#' })
#'
#' if(interactive())
#'  app$start()
#'
#' @name responses
#' @return An Ambiorix response.
#' @export
response <- function(body, headers = list(), status = 200L) {
  assert_that(not_missing(body))
  res <- list(
    status = as.integer(status),
    headers = headers,
    body = convert_body(body)
  )
  construct_response(res)
}

#' @rdname responses
#' @export
response_404 <- function(
  body = "404: Not found",
  headers = list("Content-Type" = content_html()),
  status = 404L
) {
  res <- list(
    status = as.integer(status),
    headers = headers,
    body = convert_body(body)
  )
  construct_response(res)
}

#' @rdname responses
#' @export
response_500 <- function(
  body = "500: Server Error",
  headers = list("Content-Type" = content_html()),
  status = 500L
) {
  res <- list(
    status = as.integer(status),
    headers = headers,
    body = convert_body(body)
  )
  construct_response(res)
}

#' Convert body
#'
#' Body may only be a character vector of length 1.
#'
#' @param body Body of response.
#'
#' @keywords internal
#' @noRd
convert_body <- function(body) {
  UseMethod("convert_body")
}

#' @export
convert_body.default <- function(body) {
  # by default, do not force conversion to character. see
  # https://github.com/ambiorix-web/ambiorix/issues/44
  body
}

#' @export
convert_body.factor <- function(body) {
  as.character(body)
}

#' @export
convert_body.shiny.tag <- function(body) {
  render_htmltools(body)
}

#' @export
convert_body.shiny.tag.list <- function(body) {
  render_htmltools(body)
}

#' Construct Response
#'
#' @noRd
#' @keywords internal
construct_response <- function(res) {
  structure(res, class = c(class(res), "ambiorixResponse"))
}

inline_dependencies <- function(deps) {
  lapply(
    X = deps,
    FUN = function(dep) {
      if (!length(dep$src) || !length(dep$src$file)) {
        return()
      }

      f <- function(file_name, type = c("text/css", "application/javascript")) {
        type <- match.arg(arg = type)
        tag <- switch(
          EXPR = type,
          "text/css" = htmltools::tags$style,
          "application/javascript" = htmltools::tags$script
        )

        content <- paste0(
          read_lines(file.path(dep$src$file, file_name)),
          collapse = "\n"
        )

        tag(type = type, htmltools::HTML(content))
      }

      scripts <- lapply(
        X = dep$script,
        FUN = f,
        type = "application/javascript"
      )
      styles <- lapply(X = dep$stylesheet, FUN = f, type = "text/css")

      list(scripts, styles)
    }
  )
}

render_htmltools <- function(x) {
  # if it has a <html> tag we assume
  # it's a document and render with
  # dependencies, etc.
  # otherwise we just render the tags.
  q <- htmltools::tagQuery(x)

  if (!length(q$closest("html")$selectedTags())) {
    return(htmltools::renderTags(x)$html)
  }

  deps <- htmltools::resolveDependencies(
    dependencies = htmltools::findDependencies(x)
  )
  inline_deps <- inline_dependencies(deps)

  # htmltools::renderDependencies(..., srcType = "href")
  # does not work
  rendered_deps <- htmltools::renderDependencies(deps)
  href_deps <- grep("http", strsplit(rendered_deps, "\n")[[1]], value = TRUE)
  href_deps <- paste0(href_deps, collapse = "\n")

  # add <body> if not present, and enclose all children tags within it, <head> tags will
  # be extracted thanks to htmltools::renderTags
  if (!length(q$find("body")$selectedTags())) {
    q$closest("html")$empty()$append(htmltools::tags$body(x$children))
  }

  # add <head> if not present
  if (!length(q$find("head")$selectedTags())) {
    q$closest("html")$prepend(htmltools::tags$head())
  }

  # add encoding and dependencies for the first selected tag this avoid duplicates as
  # append *appends* for each selected tag
  q$closest("html")$find("head")$filter(
    function(x, i) {
      i == 1
    }
  )$prepend(
    htmltools::tags$meta(charset = "UTF-8")
  )$append(
    htmltools::HTML(href_deps),
    inline_deps
  )

  # add placeholder for head tag children
  q$closest("html")$prepend(htmltools::HTML(
    "<head>\n<!--HEAD_CONTENT-->\n</head>"
  ))
  # get all tags and render
  x <- q$allTags()
  rendered <- htmltools::renderTags(x)

  paste0(
    "<!DOCTYPE html>\n",
    sub(
      pattern = "<!--HEAD_CONTENT-->",
      replacement = rendered$head,
      x = rendered$html,
      fixed = TRUE
    )
  )
}

#' @export
print.ambiorixResponse <- function(x, ...) {
  message("An ambiorix response")
}

#' @keywords internal
is_response <- function(obj) {
  inherits(obj, "ambiorixResponse")
}

#' Response
#'
#' Response class to generate responses sent from the server.
#'
#' @field status Status of the response, defaults to `200L`.
#' @field headers Named list of headers.
#'
#' @return A Response object.
#' @examples
#' if (interactive()) {
#'   library(ambiorix)
#'
#'   app <- Ambiorix$new()
#'
#'   app$get("/", function(req, res) {
#'     # print(res)
#'     res$send("Using {ambiorix}!")
#'   })
#'
#'   app$start()
#' }
#' @export
Response <- R6::R6Class(
  "Response",
  lock_objects = FALSE,
  public = list(
    #' @details Set the status of the response.
    #' @param status An integer defining the status.
    set_status = function(status) {
      assert_that(not_missing(status))
      private$.status <- status
      invisible(self)
    },
    #' @details Send a plain HTML response.
    #' @param body Body of the response.
    send = function(body) {
      response(
        body = body,
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Send a plain HTML response, pre-processed with sprintf.
    #' @param body Body of the response.
    #' @param ... Passed to `...` of `sprintf`.
    sendf = function(body, ...) {
      response(
        body = sprintf(body, ...),
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Send a plain text response.
    #' @param body Body of the response.
    text = function(body) {
      self$header_content_plain()

      response(
        body = body,
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Send a file.
    #' @param file File to send.
    send_file = function(file) {
      assert_that(not_missing(file))

      body <- readBin(
        con = file,
        what = "raw",
        n = file.info(file)$size
      )

      self$header(
        name = "Content-Type",
        value = mime::guess_type(file = file)
      )

      response(
        body = body,
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Redirect to a path or URL.
    #' @param path Path or URL to redirect to.
    redirect = function(path) {
      status <- private$.get_status()

      # for redirection, must start with a "3". if not, default
      # to 302. see #138.
      if (!startsWith(x = as.character(status), prefix = "3")) {
        status <- 302L
      }

      self$header(
        name = "Location",
        value = path
      )

      response(
        body = "",
        headers = private$.get_headers(),
        status = status
      )
    },
    #' @details Render a template file.
    #' @param file Template file.
    #' @param data List to fill `[% tags %]`.
    render = function(file, data = list()) {
      assert_that(not_missing(file))
      assert_that(has_file(file))

      response(
        body = private$.render_template(file, data),
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Render an object as JSON.
    #' @param body Body of the response.
    #' @param ... Additional named arguments passed to the serialiser.
    json = function(body, ...) {
      self$header_content_json()

      response(
        body = serialise(body, ...),
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Sends a comma separated value file
    #' @param data Data to convert to CSV.
    #' @param name Name of the file.
    #' @param ... Additional arguments passed to [readr::format_csv()].
    csv = function(data, name = "data", ...) {
      assert_that(not_missing(data))
      check_installed("readr")

      name <- sprintf("attachment;charset=UTF-8;filename=%s.csv", name)
      self$header(
        name = "Content-Disposition",
        value = name
      )
      self$header_content_csv()

      data <- readr::format_csv(x = data, ...)

      response(
        body = data,
        header = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Sends a tab separated value file
    #' @param data Data to convert to CSV.
    #' @param name Name of the file.
    #' @param ... Additional arguments passed to [readr::format_tsv()].
    tsv = function(data, name = "data", ...) {
      assert_that(not_missing(data))
      check_installed("readr")

      name <- sprintf("attachment;charset=UTF-8;filename=%s.tsv", name)
      self$header(
        name = "Content-Disposition",
        value = name
      )
      self$header_content_tsv()

      data <- readr::format_tsv(x = data, ...)

      response(
        body = data,
        header = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Sends an htmlwidget.
    #' @param widget The widget to use.
    #' @param ... Additional arguments passed to [htmlwidgets::saveWidget()].
    htmlwidget = function(widget, ...) {
      check_installed("htmlwidgets")
      if (!inherits(widget, "htmlwidget")) {
        stop("This is not an htmlwidget", call. = FALSE)
      }

      # save and read
      tmp <- tempfile(fileext = ".html")
      htmlwidgets::saveWidget(widget, tmp, selfcontained = TRUE, ...)
      on.exit(expr = unlink(tmp))

      response(
        body = paste0(read_lines(tmp), "\n", collapse = ""),
        headers = private$.get_headers(),
        status = private$.get_status()
      )
    },
    #' @details Render a markdown file.
    #' @param file Template file.
    #' @param data List to fill `[% tags %]`.
    md = function(file, data = list()) {
      check_installed("commonmark")
      self$render(
        file = file,
        data = data
      )
    },
    #' @details Send a png file
    #' @param file Path to local file.
    png = function(file) {
      private$.send_image(file, "png")
    },
    #' @details Send a jpeg file
    #' @param file Path to local file.
    jpeg = function(file) {
      private$.send_image(file, "jpeg")
    },
    #' @details Send an image
    #' Similar to `png` and `jpeg` methods but guesses correct method
    #' based on file extension.
    #' @param file Path to local file.
    image = function(file) {
      type <- tools::file_ext(file)
      if (!type %in% c("png", "jpeg")) {
        stop("Only accepts .png and .jpeg files")
      }
      private$.send_image(file, type)
    },
    #' @details Ggplot2
    #' @param plot Ggplot2 plot object.
    #' @param type Type of image to save.
    #' @param ... Passed to [ggplot2::ggsave()]
    ggplot2 = function(plot, ..., type = c("png", "jpeg")) {
      assert_that(not_missing(plot))
      check_installed("ggplot2")

      type <- match.arg(type)
      ext <- sprintf(".%s", type)
      temp <- tempfile(fileext = ext)
      ggplot2::ggsave(
        temp,
        plot,
        ...
      )
      private$.send_image(temp, type, clean = TRUE)
    },
    #' @details Print
    print = function() {
      cli::cli_h3("A Response")

      if (!length(private$.headers)) {
        return(invisible())
      }

      cli::cli_h3("Headers")
      cli::cli_ul()

      for (i in seq_along(private$.headers)) {
        cli::cli_li("HEADER {names(private$.headers)[i]}")
        str(private$.headers[[i]])
      }

      cli::cli_end()
    },
    #' @details Add headers to the response.
    #' @param name String. Name of the header.
    #' @param value Value of the header.
    #' @return Invisibly returns self.
    header = function(name, value) {
      assert_that(not_missing(name))
      assert_that(not_missing(value))
      private$.headers[[name]] <- value
      invisible(self)
    },
    #' @details Set Content Type to JSON
    #' @return Invisibly returns self.
    header_content_json = function() {
      self$header("Content-Type", content_json())
      invisible(self)
    },
    #' @details Set Content Type to HTML
    #' @return Invisibly returns self.
    header_content_html = function() {
      self$header("Content-Type", content_html())
      invisible(self)
    },
    #' @details Set Content Type to Plain Text
    #' @return Invisibly returns self.
    header_content_plain = function() {
      self$header("Content-Type", content_plain())
      invisible(self)
    },
    #' @details Set Content Type to CSV
    #' @return Invisibly returns self.
    header_content_csv = function() {
      self$header("Content-Type", content_csv())
      invisible(self)
    },
    #' @details Set Content Type to TSV
    #' @return Invisibly returns self.
    header_content_tsv = function() {
      self$header("Content-Type", content_tsv())
      invisible(self)
    },
    #' @details Get headers
    #' Returns the list of headers currently set.
    get_headers = function() {
      private$.headers
    },
    #' @details Get a header
    #' Returns a single header currently, `NULL` if not set.
    #' @param name Name of the header to return.
    get_header = function(name) {
      assert_that(not_missing(name))
      private$.headers[[name]]
    },
    #' @details Set headers
    #' @param headers A named list of headers to set.
    set_headers = function(headers) {
      assert_that(not_missing(headers))
      if (!is.list(headers)) {
        stop("`headers` must be a named list")
      }

      private$.headers <- headers
      invisible(self)
    },
    #' @details Add a pre render hook.
    #' Runs before the `render` method.
    #'
    #' @param hook A function that accepts at least 4 arguments:
    #' - `self`: The `Request` class instance.
    #' - `content`: File content a vector of character string,
    #' content of the template.
    #' - `data`: `list` passed from `render` method.
    #' - `ext`: File extension of the template file.
    #'
    #' This function is used to add pre-render hooks to the `render`
    #' method. The function should return an object of class
    #' `responsePreHook` as obtained by [pre_hook()].
    #' This is meant to be used by middlewares to, if necessary,
    #' pre-process rendered data.
    #'
    #' Include `...` in your `hook` to ensure it will handle
    #' potential updates to hooks in the future.
    #'
    #' @return Invisible returns self.
    pre_render_hook = function(hook) {
      assert_that(not_missing(hook))
      assert_that(
        is.function(hook),
        msg = "`hook` must be a function"
      )

      assert_that(
        length(formalArgs(hook)) >= 4,
        msg = "`hook` must take at least 4 arguments: `self`, `content`, `data`, and `ext`"
      )

      private$.preHooks <- append(private$.preHooks, hook)
      invisible(self)
    },
    #' @details Post render hook.
    #'
    #' @param hook A function to run after the rendering of HTML.
    #' It should accept at least 3 arguments:
    #' - `self`: The `Response` class instance.
    #' - `content`: File content a vector of character string,
    #' content of the template.
    #' - `ext`: File extension of the template file.
    #'
    #' Include `...` in your `hook` to ensure it will handle
    #' potential updates to hooks in the future.
    #'
    #' @return Invisible returns self.
    post_render_hook = function(hook) {
      assert_that(not_missing(hook))
      assert_that(
        is.function(hook),
        msg = "`hook` must be a function"
      )

      assert_that(
        length(formalArgs(hook)) >= 3,
        msg = "`hook` must take 2 arguments: `self`, `content`, and `ext`"
      )

      private$.postHooks <- append(private$.postHooks, hook)
      invisible(self)
    },
    #' @details Set a cookie
    #' Overwrites existing cookie of the same `name`.
    #' @param name String. Name of the cookie.
    #' @param value value of the cookie.
    #' @param expires Expiry, if an integer assumes it's the number of seconds
    #' from now. Otherwise accepts an object of class `POSIXct` or `Date`.
    #' If a `character` string then it is set as-is and not pre-processed.
    #' If unspecified, the cookie becomes a session cookie. A session finishes
    #' when the client shuts down, after which the session cookie is removed.
    #' @param max_age Indicates the number of seconds until the cookie expires.
    #' A zero or negative number will expire the cookie immediately.
    #' If both `expires` and `max_age` are set, the latter has precedence.
    #' @param domain Defines the host to which the cookie will be sent.
    #' If omitted, this attribute defaults to the host of the current document URL,
    #' not including subdomains.
    #' @param path Indicates the path that must exist in the requested URL for the
    #' browser to send the Cookie header.
    #' @param secure Indicates that the cookie is sent to the server only when a
    #' request is made with the https: scheme (except on localhost), and therefore,
    #' is more resistant to man-in-the-middle attacks.
    #' @param http_only Forbids JavaScript from accessing the cookie, for example,
    #' through the document.cookie property.
    #' @param same_site Controls whether or not a cookie is sent with cross-origin
    #' requests, providing some protection against cross-site request forgery
    #' attacks (CSRF). Accepts `Strict`, `Lax`, or `None`.
    #' @return Invisibly returns self.
    cookie = function(
      name,
      value,
      expires = getOption("ambiorix.cookie.expire"),
      max_age = getOption("ambiorix.cookie.maxage"),
      domain = getOption("ambiorix.cookie.domain"),
      path = getOption("ambiorix.cookie.path", "/"),
      secure = getOption("ambiorix.cookie.secure", TRUE),
      http_only = getOption("ambiorix.cookie.httponly", TRUE),
      same_site = getOption("ambiorix.cookie.savesite")
    ) {
      assert_that(not_missing(name))
      assert_that(not_missing(value))

      if (length(.globals$cookiePreprocessors) > 0) {
        for (i in seq_along(.globals$cookiePreprocessors)) {
          value <- .globals$cookiePreprocessors[[i]](
            name,
            value,
            expires,
            max_age,
            domain,
            path,
            secure,
            http_only,
            same_site
          )
        }
      }

      private$.cookies[[name]] <- cookie(
        name,
        value,
        expires,
        max_age,
        domain,
        path,
        secure,
        http_only,
        same_site
      )

      invisible(self)
    },
    #' @details Clear a cookie
    #' Clears the value of a cookie.
    #' @param name Name of the cookie to clear.
    #' @return Invisibly returns self.
    clear_cookie = function(name) {
      # cookies with date in the past are removed from the browser
      self$cookie(
        name,
        "",
        expires = Sys.Date() - 365L
      )

      invisible(self)
    }
  ),
  active = list(
    status = function(value) {
      if (missing(value)) {
        return(private$.status)
      }

      value <- as.integer(value)
      is_valid_status <- length(value) == 1L &&
        !is.na(value) &&
        is.integer(value)

      if (!is_valid_status) {
        msg <- paste0("Invalid response status: '", value, "'")
        stop(msg, call. = FALSE)
      }

      private$.status <- value
    },
    headers = function(value) {
      if (missing(value)) {
        return(private$.headers)
      }

      if (!is.list(value)) {
        stop("Must be a `list`")
      }

      private$.headers <- value
    }
  ),
  private = list(
    .status = 200L,
    .cookies = list(),
    .headers = list(),
    .preHooks = list(),
    .postHooks = list(),
    .render_template = function(file, data) {
      file <- normalizePath(file)

      if (!is.null(.globals$renderer)) {
        response <- .globals$renderer(file, data)
        # if the response is NULL we keep going
        if (isTRUE(!is.null(response))) {
          return(response)
        }
      }

      # read and replace tags
      file_content <- read_lines_cached(file)

      # render
      ext <- tools::file_ext(file)

      # handle partials
      # replace brackets so glue::glue_data evals
      file_content <- replace_partials(file_content, get_dir(file))

      if (ext == "html") {
        data <- lapply(data, function(x) {
          if (!inherits(x, "jobj")) {
            return(x)
          }

          as.character(serialise(x))
        })
      }

      # hooks
      if (length(private$.preHooks) > 0) {
        for (i in seq_along(private$.preHooks)) {
          pre_processed <- private$.preHooks[[i]](self, file_content, data, ext)
          if (!is_pre_hook(pre_processed)) {
            message(
              error(),
              "Not a valid return value from pre-hook (ignoring)"
            )
            next
          }

          file_content <- pre_processed$content
          data <- pre_processed$data
        }
      }

      file_content <- render_tags(file_content, data)

      if (ext == "md") {
        file_content <- commonmark::markdown_html(file_content)
      }

      # collapse html
      if (ext == "html" || ext == "md") {
        return(private$.run_post_hooks(
          paste0(file_content, collapse = ""),
          ext
        ))
      }

      # parse R
      private$.run_post_hooks(render_html(file_content), ext)
    },
    .get_status = function(status = NULL) {
      if (is.null(status)) {
        return(private$.status)
      }

      status
    },
    .get_headers = function(headers = NULL) {
      private$.render_cookies()
      heads <- private$.headers

      if (!is.null(headers)) {
        heads <- modifyList(heads, headers)
      }

      if (is.null(heads[["Content-Type"]])) {
        heads <- modifyList(heads, list("Content-Type" = content_html()))
      }

      heads
    },
    .run_post_hooks = function(file_content, ext) {
      if (length(private$.postHooks) == 0) {
        return(file_content)
      }

      for (i in seq_along(private$.postHooks)) {
        content <- private$.postHooks[[i]](self, file_content, ext)

        if (!is.character(content)) {
          message(
            error(),
            "Not a character vector returned from post-hook (ignoring)"
          )
          next
        }

        file_content <- content
      }

      file_content
    },
    .render_cookies = function() {
      if (length(private$.cookies) == 0L) {
        return()
      }

      for (opts in private$.cookies) {
        cookie <- sprintf("%s=%s", opts$name, opts$value)

        if (!is.null(opts$expires)) {
          expires <- convert_cookie_expires(opts$expires)
          cookie <- sprintf("%s; Expires=%s", cookie, expires)
        }

        if (!is.null(opts$max_age)) {
          cookie <- sprintf("%s; Max-Age=%s", cookie, opts$max_age)
        }

        if (!is.null(opts$domain)) {
          cookie <- sprintf("%s; Domain=%s", cookie, opts$domain)
        }

        if (!is.null(opts$path)) {
          cookie <- sprintf("%s; Path=%s", cookie, opts$path)
        }

        if (opts$secure) {
          cookie <- sprintf("%s; Secure", cookie)
        }

        if (opts$http_only) {
          cookie <- sprintf("%s; HttpOnly", cookie)
        }

        if (!is.null(opts$same_site)) {
          cookie <- sprintf("%s; SameSite=%s", cookie, opts$same_site)
        }

        names(cookie) <- "Set-Cookie"

        private$.headers <- append(
          private$.headers,
          as.list(cookie)
        )
      }
    },
    .send_image = function(file, type = c("png", "jpeg"), clean = FALSE) {
      assert_that(not_missing(file))

      if (grepl("http", file)) {
        stop("Must be a local file", call. = FALSE)
      }

      if (clean) {
        on.exit({
          unlink(file)
        })
      }

      type <- match.arg(type)
      type <- sprintf("image/%s", type)

      size <- file.info(file)$size
      con <- file(file, "rb")
      on.exit(
        {
          close(con)
        },
        add = TRUE
      )

      raw <- readBin(con, raw(), size)

      self$header("Content-Length", size)
      self$header("Content-Type", type)
      self$send(raw)
    }
  )
)

#' Convert Cookie Expires
#'
#' Converts the cookie `expires` argument
#' to the expected
#' [Date format](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Date).
#'
#' @param expires Expiry, if an integer assumes it's the number of seconds
#' from now. Otherwise accepts an object of class `POSIXct` or `Date`.
#'
#' @examples
#' # expires in an hour
#' convert_cookie_expires(60 * 60)
#'
#' # expires tomorrow
#' convert_cookie_expires(Sys.Date() + 1)
#'
#' # expires in 1 minute
#' convert_cookie_expires(Sys.time() + 60)
#'
#' @noRd
#' @keywords internal
convert_cookie_expires <- function(expires) {
  if (is.character(expires)) {
    return(expires)
  }

  if (is.numeric(expires)) {
    expires <- as.POSIXct(Sys.time(), tz = "UTC") + expires
  }

  if (inherits(expires, "Date") || inherits(expires, "POSIXct")) {
    expires <- as.POSIXct(expires, tz = "UTC")
    expires <- format(
      expires,
      "%a, %d %b %Y %H:%M:%S GMT"
    )
  }

  expires
}

#' Log Predicate
#'
#' Log predicate to bridge the previous log with the new
#' [log::Logger] package.
#'
#' @noRd
#' @keywords internal
logPredicate <- function(log) {
  function() log
}

#' Logger
#'
#' Returns a new logger using the `log` package.
#'
#' @param prefix String to prefix all log messages.
#' @param write Whether to write the log to the `file`.
#' @param file Name of the file to dump the logs to,
#' only used if `write` is `TRUE`.
#' @param sep Separator between `prefix` and other
#' flags and messages.
#'
#' @examples
#' log <- new_log()
#' log$log("Hello world")
#'
#' @return An R& of class `log::Logger`.
#'
#' @export
new_log <- function(
  prefix = ">",
  write = FALSE,
  file = "ambiorix.log",
  sep = ""
) {
  log::Logger$new(
    prefix = prefix,
    write = write,
    file = file,
    sep = sep
  )$date()$time()
}

#' Customise logs
#'
#' Customise the internal logs used by Ambiorix.
#'
#' @param log An object of class `Logger`, see
#' [log::Logger].
#'
#' @name set_log
#' @return The `log` object.
#' @examples
#' # define custom loggers:
#' info_logger <- log::Logger$new("INFO")
#' success_logger <- log::Logger$new("SUCCESS")
#' error_logger <- log::Logger$new("ERROR")
#'
#' info_logger$log("This is an info message.")
#' success_logger$log("This is a success message.")
#' error_logger$log("This is an error message.")
#'
#' # set custom loggers for Ambiorix:
#' set_log_info(info_logger)
#' set_log_success(success_logger)
#' set_log_error(error_logger)
#' @export
set_log_info <- function(log) {
  assert_that(not_missing(log))
  assert_that(is_logger(log))

  .globals$infoLog <- log
}

#' @rdname set_log
#' @export
set_log_success <- function(log) {
  assert_that(not_missing(log))
  assert_that(is_logger(log))

  .globals$successLog <- log
}

#' @rdname set_log
#' @export
set_log_error <- function(log) {
  assert_that(not_missing(log))
  assert_that(is_logger(log))

  .globals$errorLog <- log
}

#' CLI symbols
#'
#' CLI Symbols for log
#'
#' @keywords internal
#' @noRd
success <- function() {
  cli::col_green(cli::symbol$tick)
}

#' @keywords internal
#' @noRd
error <- function() {
  cli::col_red(cli::symbol$cross)
}

#' @keywords internal
#' @noRd
info <- function() {
  cli::col_blue(cli::symbol$info)
}

#' @keywords internal
#' @noRd
warn <- function() {
  cli::col_yellow(cli::symbol$warning)
}

test_that("log", {
  l <- new_log()
  l$log("hello")

  expect_snapshot(success())
  expect_snapshot(error())
  expect_snapshot(info())
  expect_snapshot(warn())

  set_log_error(l)
  set_log_success(l)
  set_log_info(l)
  expect_equal(.globals$successLog, l)
  expect_equal(.globals$infoLog, l)
  expect_equal(.globals$errorLog, l)
})

#' Browse App
#'
#' Browses the application, if RStudio available uses pane.
#'
#' @param open Whether to open the app.
#' @param url URL to browse.
#'
#' @noRd
#' @keywords internal
browse_ambiorix <- function(open, url) {
  if (!open) {
    return()
  }

  viewer <- getOption("viewer", browseURL)

  x <- tryCatch(
    viewer(url),
    error = function(e) e
  )

  if (inherits(x, "error")) {
    message("Unable to open browser, please open manually.")
  }

  invisible()
}

`%response%` <- function(lhs, rhs) {
  if (is.null(lhs)) {
    return(rhs)
  }
  if (!inherits(lhs, "ambiorixResponse")) {
    return(rhs)
  }

  lhs
}

`%error%` <- function(lhs, rhs) {
  if (is.null(lhs)) {
    return(rhs)
  }

  lhs
}

#' Remove Extensions
#'
#' Remove extensions from files.
#'
#' @noRd
#' @keywords internal
remove_extensions <- function(files) {
  tools::file_path_sans_ext(files)
}

#' Checks if Package is Installed
#'
#' Checks if a package is installed, stops if not.
#'
#' @param pkg Package to check.
#'
#' @noRd
#' @keywords internal
check_installed <- function(pkg) {
  has_it <- base::requireNamespace(pkg, quietly = TRUE)

  if (!has_it) {
    stop(sprintf("This function requires the package {%s}", pkg), call. = FALSE)
  }
}

#' Retrieve the Port for Server Binding
#'
#' Determines the appropriate port for the server, following a
#' specific order of precedence.
#'
#' @param host String. The host address to bind to when selecting a random port.
#' @param port Integer. An optional input port, provided by the user.
#' @details The port to use is resolved in the following order of precedence:
#' 1. Forced Port Option: Typically used by Belgic, the load balancer. If
#'   the `ambiorix.port.force` R option is specified, this port is returned.
#'   This option should not be altered during development.
#' 2. `AMBIORIX_PORT` environment variable
#' 3. `port` argument.
#' 4. `SHINY_PORT` environment variable.
#' 5. Random port.
#'
#' @return Integer. Port value to bind the server.
#'
#' @noRd
#' @keywords internal
get_port <- function(host, port = NULL) {
  # we need to override the port if the load balancer
  # is running. This should NOT be set by a dev
  # this ensures we can overwrite
  forced <- getOption("ambiorix.port.force")
  if (!is.null(forced)) {
    return(forced)
  }

  has_ints_only <- function(x) !grepl(pattern = "\\D", x = x)
  is_valid_port <- function(x) {
    !is.null(x) && !identical(x, "") && has_ints_only(x)
  }

  ambiorix_port <- Sys.getenv("AMBIORIX_PORT")
  if (is_valid_port(ambiorix_port)) {
    return(as.integer(ambiorix_port))
  }

  if (is_valid_port(port)) {
    return(as.integer(port))
  }

  shiny_port <- Sys.getenv("SHINY_PORT")
  if (is_valid_port(shiny_port)) {
    return(as.integer(shiny_port))
  }

  httpuv::randomPort(host = host)
}

#' Silent readLines
#'
#' Avoids EOF warnings.
#'
#' @param ... Passed to [readLines()]:
#'
#' @keywords internal
#' @noRd
read_lines <- function(...) {
  readLines(..., warn = FALSE)
}

#' Read file from disk or cache
#'
#' @param path Path to file.
#'
#' @keywords internal
#' @noRd
read_lines_cached <- function(path) {
  if (!.globals$cache_tmpls) {
    return(read_lines(path))
  }

  if (length(.cache_tmpls[[path]]) > 0L) {
    return(.cache_tmpls[[path]])
  }

  content <- read_lines(path)
  .cache_tmpls[[path]] <- content

  content
}

#' Ignore Unused Imports
#'
#' R CMD check does not detect imports in R6 classes
#' because it does not consider code that runs at install
#' time.
#'
#' Having this dummy function helps avoid this note:
#' "Namespaces in Imports field not imported from"
#'
#' @keywords internal
#' @noRd
ignore_unused_imports <- function() {
  mime::guess_type
  promises::promise
}

#' Token
#'
#' Create a token
#'
#' @param n Number of bytes.
#' @examples
#' token_create()
#' token_create(n = 32L)
#' @return Length 1 character vector.
#' @export
token_create <- function(n = 16L) {
  paste(
    as.hexmode(
      sample(
        256,
        n,
        replace = TRUE
      )
    ),
    collapse = ""
  )
}

Route <- R6::R6Class(
  "Route",
  public = list(
    path = NULL,
    components = list(),
    pattern = NULL,
    dynamic = FALSE,
    params = NULL,
    basepath = NULL,
    initialize = function(path) {
      assert_that(not_missing(path))
      self$path <- gsub("\\?.*$", "", path) # remove query
      self$dynamic <- grepl(":", path)
    },
    as_pattern = function(parent = "") {
      if (!is.null(.globals$pathToPattern)) {
        self$pattern <- .globals$pathToPattern(self$path)
        return(
          invisible(self)
        )
      }

      pattern <- sapply(self$components, function(comp) {
        if (comp$dynamic) {
          self$params <- append(self$params, comp$name)
          return("[[:alnum:][:space:][:punct:]]*")
        }

        return(comp$name)
      })

      pattern <- paste0(pattern, collapse = "/")
      self$pattern <- paste0("^", parent, "/", pattern, "$")
      invisible(self)
    },
    decompose = function(parent = "") {
      path <- paste0(parent, self$path)
      # split
      components <- strsplit(path, "(?<=.)(?=[:/])", perl = TRUE)[[1]]

      # remove lonely /
      components <- components[components != "/"]

      if (length(components) == 0) {
        self$components <- list(
          list(
            index = 1L,
            dynamic = FALSE,
            name = ""
          )
        )
        return()
      }

      # cleanup
      components <- gsub("/", "", components)

      components <- as.list(components)
      comp <- list()
      for (i in seq_along(components)) {
        c <- list(
          index = i,
          dynamic = grepl(":", components[[i]]),
          name = gsub(":|$", "", components[[i]])
        )
        comp <- append(comp, list(c))
      }

      self$components <- comp
      invisible(self)
    },
    print = function() {
      cli::cli_rule("Ambiorix", right = "route")
      message("Only used internally")
    }
  )
)

#' Path to pattern
#'
#' Identify a function as a path to pattern function;
#' a function that accepts a path and returns a matching pattern.
#'
#' @param path A function that accepts a character vector of length 1
#' and returns another character vector of length 1.
#'
#' @examples
#' fn <- function(path) {
#'   pattern <- gsub(":([^/]+)", "(\\\\w+)", path)
#'   paste0("^", pattern, "$")
#' }
#'
#' path_to_pattern <- as_path_to_pattern(fn)
#'
#' path <- "/dashboard/profile/:user_id"
#' pattern <- path_to_pattern(path) # "^/dashboard/profile/(\\w+)$"
#'
#' @return Object of class "pathToPattern".
#' @export
as_path_to_pattern <- function(path) {
  assert_that(is_function(path))

  structure(
    path,
    class = c(
      "pathToPattern",
      class(path)
    )
  )
}


#' @export
print.pathToPattern <- function(x, ...) {
  cli::cli_alert_info("A path to pattern converter")
}

#' @keywords internal
#' @noRd
is_path_to_pattern <- function(obj) {
  inherits(obj, "pathToPattern")
}

#' Request
#'
#' A request.
#'
#' @field HEADERS Headers from the request.
#' @field HTTP_ACCEPT Content types to accept.
#' @field HTTP_ACCEPT_ENCODING Encoding of the request.
#' @field HTTP_ACCEPT_LANGUAGE Language of the request.
#' @field HTTP_CACHE_CONTROL Directorives for the cache (case-insensitive).
#' @field HTTP_CONNECTION Controls whether the network connection stays open after the current transaction finishes.
#' @field HTTP_COOKIE Cookie data.
#' @field HTTP_HOST Host making the request.
#' @field HTTP_SEC_FETCH_DEST Indicates the request's destination. That is the initiator of the original fetch request, which is where (and how) the fetched data will be used.
#' @field HTTP_SEC_FETCH_MODE Indicates mode of the request.
#' @field HTTP_SEC_FETCH_SITE Indicates the relationship between a request initiator's origin and the origin of the requested resource.
#' @field HTTP_SEC_FETCH_USER Only sent for requests initiated by user activation, and its value will always be `?1`.
#' @field HTTP_UPGRADE_INSECURE_REQUESTS Signals that server supports upgrade.
#' @field HTTP_USER_AGENT User agent.
#' @field SERVER_NAME Name of the server.
#' @field httpuv.version Version of httpuv.
#' @field PATH_INFO Path of the request.
#' @field QUERY_STRING Query string of the request.
#' @field REMOTE_ADDR Remote address.
#' @field REMOTE_PORT Remote port.
#' @field REQUEST_METHOD Method of the request, e.g.: `GET`.
#' @field rook.errors Errors from rook.
#' @field rook.input Rook inputs.
#' @field rook.url_scheme Rook url scheme.
#' @field rook.version Rook version.
#' @field SCRIPT_NAME The initial portion of the request URL's "path" that corresponds to the application object, so that the application knows its virtual "location".  #' @field SERVER_NAME Server name.
#' @field SERVER_PORT Server port
#' @field CONTENT_LENGTH Size of the message body.
#' @field CONTENT_TYPE Type of content of the request.
#' @field HTTP_REFERER Contains an absolute or partial address of the page that makes the request.
#' @field body Request, an environment.
#' @field query Parsed `QUERY_STRING`, `list`.
#' @field params A `list` of parameters.
#' @field cookie Parsed `HTTP_COOKIE`.
#'
#' @return A Request object.
#' @examples
#' if (interactive()) {
#'   library(ambiorix)
#'
#'   app <- Ambiorix$new()
#'
#'   app$get("/", function(req, res) {
#'     print(req)
#'     res$send("Using {ambiorix}!")
#'   })
#'
#'   app$start()
#' }
#' @export
Request <- R6::R6Class(
  "Request",
  lock_objects = FALSE,
  public = list(
    HEADERS = NULL,
    HTTP_ACCEPT = NULL,
    HTTP_ACCEPT_ENCODING = NULL,
    HTTP_ACCEPT_LANGUAGE = NULL,
    HTTP_CACHE_CONTROL = NULL,
    HTTP_CONNECTION = NULL,
    HTTP_COOKIE = NULL,
    HTTP_HOST = NULL,
    HTTP_SEC_FETCH_DEST = NULL,
    HTTP_SEC_FETCH_MODE = NULL,
    HTTP_SEC_FETCH_SITE = NULL,
    HTTP_SEC_FETCH_USER = NULL,
    HTTP_UPGRADE_INSECURE_REQUESTS = NULL,
    HTTP_USER_AGENT = NULL,
    httpuv.version = NULL,
    PATH_INFO = NULL,
    QUERY_STRING = NULL,
    REMOTE_ADDR = NULL,
    REMOTE_PORT = NULL,
    REQUEST_METHOD = NULL,
    rook.errors = NULL,
    rook.input = NULL,
    rook.url_scheme = NULL,
    rook.version = NULL,
    SCRIPT_NAME = NULL,
    SERVER_NAME = NULL,
    SERVER_PORT = NULL,
    CONTENT_LENGTH = NULL,
    CONTENT_TYPE = NULL,
    HTTP_REFERER = NULL,
    body = NULL,
    query = list(),
    params = list(),
    cookie = list(),
    #' @details Constructor
    #' @param req Original request (environment).
    initialize = function(req) {
      self$HEADERS <- as.list(req$HEADERS)
      self$HTTP_ACCEPT <- req$HTTP_ACCEPT
      self$HTTP_ACCEPT_ENCODING <- req$HTTP_ACCEPT_ENCODING
      self$HTTP_ACCEPT_LANGUAGE <- req$HTTP_ACCEPT_LANGUAGE
      self$HTTP_CACHE_CONTROL <- req$HTTP_CACHE_CONTROL
      self$HTTP_CONNECTION <- req$HTTP_CONNECTION
      self$HTTP_COOKIE <- req$HTTP_COOKIE
      self$HTTP_HOST <- req$HTTP_HOST
      self$HTTP_SEC_FETCH_DEST <- req$HTTP_SEC_FETCH_DEST
      self$HTTP_SEC_FETCH_MODE <- req$HTTP_SEC_FETCH_MODE
      self$HTTP_SEC_FETCH_SITE <- req$HTTP_SEC_FETCH_SITE
      self$HTTP_SEC_FETCH_USER <- req$HTTP_SEC_FETCH_USER
      self$HTTP_UPGRADE_INSECURE_REQUESTS <- req$HTTP_UPGRADE_INSECURE_REQUESTS
      self$HTTP_USER_AGENT <- req$HTTP_USER_AGENT
      self$httpuv.version <- req$httpuv.version
      self$PATH_INFO <- req$PATH_INFO
      self$QUERY_STRING <- req$QUERY_STRING
      self$REMOTE_ADDR <- req$REMOTE_ADDR
      self$REMOTE_PORT <- req$REMOTE_PORT
      self$REQUEST_METHOD <- req$REQUEST_METHOD
      self$rook.errors <- req$rook.errors
      self$rook.input <- req$rook.input
      self$rook.url_scheme <- req$rook.url_scheme
      self$rook.version <- req$rook.version
      self$SCRIPT_NAME <- req$SCRIPT_NAME
      self$SERVER_NAME <- req$SERVER_NAME
      self$SERVER_PORT <- req$SERVER_NAME
      self$CONTENT_LENGTH <- req$CONTENT_LENGTH
      self$CONTENT_TYPE <- req$CONTENT_TYPE
      self$HTTP_REFERER <- req$HTTP_REFERER
      self$body <- req

      private$.parse_query_string(req$QUERY_STRING)
      self$cookie <- .globals$cookieParser(req)
    },
    #' @details Print
    print = function() {
      cli::cli_h3("A Request")
      cli::cli_ul()
      cli::cli_li("HEADERS: {.val {self$HEADERS}}")
      cli::cli_li("HTTP_ACCEPT: {.val {self$HTTP_ACCEPT}}")
      cli::cli_li("HTTP_ACCEPT_ENCODING: {.val {self$HTTP_ACCEPT_ENCODING}}")
      cli::cli_li("HTTP_ACCEPT_LANGUAGE: {.val {self$HTTP_ACCEPT_LANGUAGE}}")
      cli::cli_li("HTTP_CACHE_CONTROL: {.val {self$HTTP_CACHE_CONTROL}}")
      cli::cli_li("HTTP_CONNECTION: {.val {self$HTTP_CONNECTION}}")
      cli::cli_li("HTTP_COOKIE: {.val {self$HTTP_COOKIE}}")
      cli::cli_li("HTTP_HOST: {.val {self$HTTP_HOST}}")
      cli::cli_li("HTTP_SEC_FETCH_DEST: {.val {self$HTTP_SEC_FETCH_DEST}}")
      cli::cli_li("HTTP_SEC_FETCH_MODE: {.val {self$HTTP_SEC_FETCH_MODE}}")
      cli::cli_li("HTTP_SEC_FETCH_SITE: {.val {self$HTTP_SEC_FETCH_SITE}}")
      cli::cli_li("HTTP_SEC_FETCH_USER: {.val {self$HTTP_SEC_FETCH_USER}}")
      cli::cli_li(
        "HTTP_UPGRADE_INSECURE_REQUESTS: {.val {self$HTTP_UPGRADE_INSECURE_REQUESTS}}"
      )
      cli::cli_li("HTTP_USER_AGENT: {.val {self$HTTP_USER_AGENT}}")
      cli::cli_li("httpuv.version {.val {self$httpuv.version}}")
      cli::cli_li("PATH_INFO: {.val {self$PATH_INFO}}")
      cli::cli_li("QUERY_STRING: {.val {self$QUERY_STRING}}")
      cli::cli_li("REMOTE_ADDR: {.val {self$REMOTE_ADDR}}")
      cli::cli_li("REMOTE_PORT: {.val {self$REMOTE_PORT}}")
      cli::cli_li("REQUEST_METHOD: {.val {self$REQUEST_METHOD}}")
      cli::cli_li("SCRIPT_NAME: {.val {self$SCRIPT_NAME}}")
      cli::cli_li("SERVER_NAME: {.val {self$SERVER_NAME}}")
      cli::cli_li("SERVER_PORT: {.val {self$SERVER_PORT}}")
      cli::cli_li("CONTENT_LENGTH: {.val {self$CONTENT_LENGTH}}")
      cli::cli_li("CONTENT_TYPE: {.val {self$CONTENT_TYPE}}")
      cli::cli_li("HTTP_REFERER: {.val {self$HTTP_REFERER}}")
      cli::cli_li("rook.version: {.val {self$rook.version}}")
      cli::cli_li("rook.url_scheme: {.val {self$rook.url_scheme}}")

      if (length(self$params)) {
        cli::cli_li("params: {.val params}")
        str(self$params)
      }

      if (length(self$query)) {
        cli::cli_li("query: {.val query}")
        str(self$query)
      }

      cli::cli_end()
    },
    #' @details Get Header
    #' @param name Name of the header
    get_header = function(name) {
      assert_that(not_missing(name))
      req$HEADERS[[name]]
    },
    #' @details Parse Multipart encoded data
    parse_multipart = function() {
      parse_multipart(self)
    },
    #' @details Parse JSON encoded data
    #' @param ... Arguments passed to [parse_json()].
    parse_json = function(...) {
      parse_json(self, ...)
    }
  ),
  private = list(
    .parse_query_string = function(query) {
      if (identical(length(query), 0L)) {
        return()
      }

      self$query <- webutils::parse_query(query)
      invisible()
    }
  )
)

#' Set Parameters
#'
#' Set the query's parameters.
#'
#' @param path Corresponds the requests' `PATH_INFO`
#' @param route See `Route`
#'
#' @return Parameter list
#' @keywords internal
#' @noRd
set_params <- function(path, route = NULL) {
  if (is.null(route)) {
    return(list())
  }

  if (!route$dynamic) {
    return(list())
  }

  path_split <- strsplit(path, "/")[[1]]
  path_split <- path_split[path_split != ""]

  nms <- c()
  pms <- list()
  for (i in seq_along(path_split)) {
    if (i > length(route$components)) {
      break
    }

    if (route$components[[i]]$dynamic) {
      nms <- c(nms, route$components[[i]]$name)
      pms <- append(pms, utils::URLdecode(path_split[i]))
    }
  }

  names(pms) <- nms
  return(pms)
}

#' Mock Request
#'
#' Mock a request, used for tests.
#'
#' @param cookie Cookie string.
#' @param query Query string.
#' @param path Path string.
#'
#' @examples
#' mockRequest()
#'
#' @return A `Request` object.
#' @export
mockRequest <- function(
  cookie = "",
  query = "",
  path = "/"
) {
  req <- list(
    HEADERS = list(
      accept = "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
      `accept-encoding` = "gzip, deflate, br",
      `accept-language` = "en-US,en;q=0.9",
      connection = "keep-alive",
      host = "localhost:13698",
      `sec-ch-ua` = "\" Not A;Brand\";v=\"99\", \"Chromium\";v=\"99\"",
      `sec-ch-ua-mobile` = "?0",
      `sec-ch-ua-platform` = "\"Linux\"",
      `sec-fetch-dest` = "document",
      `sec-fetch-mode` = "navigate",
      `sec-fetch-site` = "none",
      `sec-fetch-user` = "?1",
      `upgrade-insecure-requests` = "1",
      `user-agent` = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36"
    ),
    HTTP_COOKIE = cookie,
    HTTP_ACCEPT = "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
    HTTP_ACCEPT_ENCODING = "gzip, deflate, br",
    HTTP_ACCEPT_LANGUAGE = "en-US,en;q=0.9",
    HTTP_CONNECTION = "keep-alive",
    HTTP_HOST = "localhost:13698",
    HTTP_SEC_FETCH_DEST = "document",
    HTTP_SEC_FETCH_MODE = "navigate",
    HTTP_SEC_FETCH_SITE = "none",
    HTTP_SEC_FETCH_USER = "?1",
    HTTP_UPGRADE_INSECURE_REQUESTS = "1",
    HTTP_USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36",
    httpuv.version = structure(
      list(c(1L, 6L, 5L)),
      class = c("package_version", "numeric_version")
    ),
    PATH_INFO = path,
    QUERY_STRING = query,
    REMOTE_ADDR = "127.0.0.1",
    REMOTE_PORT = "44328",
    REQUEST_METHOD = "GET",
    rook.errors = list(),
    rook.input = list(),
    rook.url_scheme = "http",
    rook.version = "1.1-0",
    SCRIPT_NAME = "",
    SERVER_NAME = "127.0.0.1",
    SERVER_PORT = "127.0.0.1"
  )

  Request$new(req)
}

WebsocketHandler <- R6::R6Class(
  "WebsocketHandler",
  public = list(
    initialize = function(name, fun) {
      assert_that(not_missing(fun))
      assert_that(not_missing(name))

      private$.fun <- fun
      private$.name <- name
    },
    receive = function(message, ws) {
      args <- formalArgs(private$.fun)

      if (length(args) > 1) {
        ws <- Websocket$new(ws)
        private$.fun(message$message, ws)
      } else {
        private$.fun(message$message)
      }
    },
    is_handler = function(message) {
      if (is.null(message$isAmbiorix)) {
        return(FALSE)
      }

      private$.name == message$name
    },
    print = function() {
      foo <- paste0(deparse(private$.fun), collapse = "\n")
      cli::cli_alert_info("receive: {.code receive(message, ws)}")
      cli::cli_ul("Listening on message:")
      cli::cli_li("name: {.val {private$.name}}")
      cli::cli_end()
    }
  ),
  private = list(
    .name = NULL,
    .fun = NULL
  )
)

#' Websocket
#'
#' Handle websocket messages.
#'
#' @return A Websocket object.
#' @examples
#' # create an Ambiorix app with websocket support:
#' if (interactive()) {
#'   library(ambiorix)
#'
#'   home_get <- function(req, res) {
#'     res$send("hello, world!")
#'   }
#'
#'   greeting_ws_handler <- function(msg, ws) {
#'     cat("Received message:", "\n")
#'     print(msg)
#'     ws$send("greeting", "Hello from the server!")
#'   }
#'
#'   app <- Ambiorix$new(port = 8080)
#'   app$get("/", home_get)
#'   app$receive("greeting", greeting_ws_handler)
#'   app$start()
#' }
#'
#' # create websocket client from another R session:
#' if (interactive()) {
#'   client <- websocket::WebSocket$new("ws://127.0.0.1:8080", autoConnect = FALSE)
#'
#'   client$onOpen(function(event) {
#'     cat("Connection opened\n")
#'
#'     msg <- list(
#'       isAmbiorix = TRUE, # __MUST__ be set!
#'       name = "greeting",
#'       message = "Hello from the client!"
#'     )
#'
#'     # serialise:
#'     msg <- yyjsonr::write_json_str(msg, auto_unbox = TRUE)
#'
#'     client$send(msg)
#'   })
#'
#'   client$onMessage(function(event) {
#'     cat("Received message from server:", event$data, "\n")
#'   })
#'
#'   client$connect()
#' }
#'
#' @export
Websocket <- R6::R6Class(
  "Websocket",
  public = list(
    #' @details Constructor
    #' @param ws The websocket
    initialize = function(ws) {
      private$.ws <- ws
    },
    #' @details Send a message
    #' @param name Name, identifier, of the message.
    #' @param message Content of the message, anything that can be
    #' serialised to JSON.
    send = function(name, message) {
      message <- list(
        name = name,
        message = message,
        isAmbiorix = TRUE
      )
      private$.ws$send(serialise(message))
    },
    #' @details Print
    print = function() {
      cli::cli_li("send: {.code send(name, message)}")
    }
  ),
  private = list(
    .ws = NULL
  )
)

#' Websocket Client
#'
#' Handle ambiorix websocket client.
#'
#' @param path Path to copy the file to.
#'
#' @section Functions:
#' - `copy_websocket_client` Copies the websocket client file, useful when
#'   ambiorix was not setup with the ambiorix generator.
#' - `get_websocket_client_path` Retrieves the full path to the local websocket client.
#' - `get_websocket_clients` Retrieves clients connected to the server.
#'
#' @name websocket_client
#' @return
#' - `copy_websocket_client`: String. The new path (invisibly).
#' - `get_websocket_client_path`: String. The path to the local websocket client.
#' - `get_websocket_clients`: List. Websocket clients.
#' @examples
#' chat_ws <- function(msg, ws) {
#'   lapply(
#'     X = get_websocket_clients(),
#'     FUN = function(c) {
#'       c$send("chat", msg)
#'     }
#'   )
#' }
#' @export
copy_websocket_client <- function(path) {
  assert_that(not_missing(path))

  lib <- get_websocket_client_path()
  file.copy(from = lib, to = path)
}

#' @rdname websocket_client
get_websocket_client_path <- function() {
  system.file("ambiorix.js", package = "ambiorix")
}

#' @rdname websocket_client
get_websocket_clients <- function() {
  return(.globals$wsc)
}

#' Forward Method
#'
#' Makes it such that the web server skips this method and uses the next one in line instead.
#'
#' @return An object of class `forward`.
#'
#' @examples
#' app <- Ambiorix$new()
#'
#' app$get("/next", function(req, res){
#'  forward()
#' })
#'
#' app$get("/next", function(req, res){
#'  res$send("Hello")
#' })
#'
#' if(interactive())
#'  app$start()
#'
#' @export
forward <- function() {
  structure("next", class = "forward")
}

#' @export
print.forward <- function(x, ...) {
  cat("Using next method")
}

#' @keywords internal
is_forward <- function(x) {
  inherits(x, "forward")
}

#' Router
#'
#' Web server.
#'
#' @field error 500 response when the route errors, must a handler function that accepts the request and the response, by default uses [response_500()].
#'
#' @examples
#' # log
#' logger <- new_log()
#' # router
#' # create router
#' router <- Router$new("/users")
#'
#' router$get("/", function(req, res){
#'  res$send("List of users")
#' })
#'
#' router$get("/:id", function(req, res){
#'  logger$log("Return user id:", req$params$id)
#'  res$send(req$params$id)
#' })
#'
#' router$get("/:id/profile", function(req, res){
#'  msg <- sprintf("This is the profile of user #%s", req$params$id)
#'  res$send(msg)
#' })
#'
#' # core app
#' app <- Ambiorix$new()
#'
#' app$get("/", function(req, res){
#'  res$send("Home!")
#' })
#'
#' # mount the router
#' app$use(router)
#'
#' if(interactive())
#'  app$start()
#'
#' @importFrom assertthat assert_that
#' @importFrom utils browseURL
#'
#' @return A Router object.
#' @export
Router <- R6::R6Class(
  "Router",
  inherit = Routing,
  public = list(
    error = NULL,
    #' @details Define the base route.
    #' @param path The base path of the router.
    initialize = function(path) {
      assert_that(not_missing(path))
      super$initialize(path)
    },
    #' @details Print
    print = function() {
      cli::cli_rule("Ambiorix", right = "router")
      cli::cli_li("routes: {.val {super$n_routes()}}")
    }
  )
)

#' Content Headers
#'
#' Convenient functions for more readable content type headers.
#'
#' @examples
#' list(
#'  "Content-Type",
#'  content_json()
#' )
#'
#' if(FALSE)
#'  req$header(
#'   "Content-Type",
#'   content_json()
#'  )
#'
#' @name content
#' @return Length 1 character vector.
#' @export
content_html <- function() {
  "text/html"
}

#' @rdname content
#' @export
content_plain <- function() {
  "text/plain"
}

#' @rdname content
#' @export
content_json <- function() {
  "application/json"
}

#' @rdname content
#' @export
content_csv <- function() {
  "text/csv"
}

#' @rdname content
#' @export
content_tsv <- function() {
  "tab-separated-values"
}

#' @rdname content
#' @export
content_protobuf <- function() {
  "application/x-protobuf"
}

test_that("content", {
  expect_equal(
    content_csv(),
    "text/csv"
  )

  expect_equal(
    content_tsv(),
    "tab-separated-values"
  )

  expect_equal(
    content_html(),
    "text/html"
  )

  expect_equal(
    content_plain(),
    "text/plain"
  )

  expect_equal(
    content_json(),
    "application/json"
  )

  expect_equal(
    content_protobuf(),
    "application/x-protobuf"
  )
})

# response

    Code
      res
    Message
      An ambiorix response

# Response

    Code
      res
    Message
      
      -- A Response 

---

    Code
      print(robj(list(x = 1L)))
    Message
      i R object
    Output
      list(x = 1L)

---

    Code
      print(jobj(list(x = 1L)))
    Message
      i JSON object
    Output
      [1] "data"

---

    Code
      res
    Message
      
      -- A Response 
      
      -- Headers 
        * HEADER Content-Type
    Output
       chr "tab-separated-values"
    Message
        * HEADER Location
    Output
       chr "/"
    Message
        * HEADER Content-Disposition
    Output
       chr "attachment;charset=UTF-8;filename=data.tsv"
    Message
        * HEADER Content-Length
    Output
       num 17645
    Message
        * HEADER xxx
    Output
       chr "hello"

# log

    Code
      success()
    Output
      <cli_ansi_string>
      [1] v

---

    Code
      error()
    Output
      <cli_ansi_string>
      [1] x

---

    Code
      info()
    Output
      <cli_ansi_string>
      [1] i

---

    Code
      warn()
    Output
      <cli_ansi_string>
      [1] !

# Websocket

    Code
      ws
    Message
      i receive: `receive(message, ws)`
        * Listening on message:
      * name: "hello"

---

    Code
      ws
    Message
      * send: `send(name, message)`

# Request

    Code
      req
    Message
      
      -- A Request 
        * HEADERS:
        text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9,
        gzip, deflate, br, en-US,en;q=0.9, keep-alive, localhost:13698, " Not
        A;Brand";v="99", "Chromium";v="99", ?0, "Linux", document, navigate, none,
        ?1, 1, and Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like
        Gecko) Chrome/99.0.4844.82 Safari/537.36
        * HTTP_ACCEPT:
        "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"
        * HTTP_ACCEPT_ENCODING: "gzip, deflate, br"
        * HTTP_ACCEPT_LANGUAGE: "en-US,en;q=0.9"
        * HTTP_CACHE_CONTROL:
        * HTTP_CONNECTION: "keep-alive"
        * HTTP_COOKIE: ""
        * HTTP_HOST: "localhost:13698"
        * HTTP_SEC_FETCH_DEST: "document"
        * HTTP_SEC_FETCH_MODE: "navigate"
        * HTTP_SEC_FETCH_SITE: "none"
        * HTTP_SEC_FETCH_USER: "?1"
        * HTTP_UPGRADE_INSECURE_REQUESTS: "1"
        * HTTP_USER_AGENT: "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36
        (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36"
        * httpuv.version 1.6.5
        * PATH_INFO: "/"
        * QUERY_STRING: ""
        * REMOTE_ADDR: "127.0.0.1"
        * REMOTE_PORT: "44328"
        * REQUEST_METHOD: "GET"
        * SCRIPT_NAME: ""
        * SERVER_NAME: "127.0.0.1"
        * SERVER_PORT: "127.0.0.1"
        * CONTENT_LENGTH:
        * CONTENT_TYPE:
        * HTTP_REFERER:
        * rook.version: "1.1-0"
        * rook.url_scheme: "http"
        * params: "params"
    Output
      List of 1
       $ x: num 1
    Message
        * query: "query"
    Output
      List of 1
       $ x: num 1

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{use_html_template}
\alias{use_html_template}
\title{HTML Template}
\usage{
use_html_template()
}
\value{
A renderer function.
}
\description{
Use \code{\link[htmltools:htmlTemplate]{htmltools::htmlTemplate()}} as renderer.
Passed to \code{use} method.
}
\examples{
use_html_template()
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/router.R \name{Router} \alias{Router} \title{Router} \value{ A Router object. } \description{ Web server. } \examples{ # log logger <- new_log() # router # create router router <- Router$new("/users") router$get("/", function(req, res){ res$send("List of users") }) router$get("/:id", function(req, res){ logger$log("Return user id:", req$params$id) res$send(req$params$id) }) router$get("/:id/profile", function(req, res){ msg <- sprintf("This is the profile of user #\%s", req$params$id) res$send(msg) }) # core app app <- Ambiorix$new() app$get("/", function(req, res){ res$send("Home!") }) # mount the router app$use(router) if(interactive()) app$start() } \section{Super class}{ \code{\link[ambiorix:Routing]{ambiorix::Routing}} -> \code{Router} } \section{Public fields}{ \if{html}{\out{
}} \describe{ \item{\code{error}}{500 response when the route errors, must a handler function that accepts the request and the response, by default uses \code{\link[=response_500]{response_500()}}.} } \if{html}{\out{
}} } \section{Methods}{ \subsection{Public methods}{ \itemize{ \item \href{#method-Router-new}{\code{Router$new()}} \item \href{#method-Router-print}{\code{Router$print()}} \item \href{#method-Router-clone}{\code{Router$clone()}} } } \if{html}{\out{
Inherited methods

- `ambiorix::Routing$engine()`
- `ambiorix::Routing$get_middleware()`
- `ambiorix::Routing$get_params()`
- `ambiorix::Routing$get_receivers()`
- `ambiorix::Routing$get_routes()`
- `ambiorix::Routing$param()`
- `ambiorix::Routing$prepare()`
- `ambiorix::Routing$receive()`
- `ambiorix::Routing$use()`

}} \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-Router-new}{}}} \subsection{Method \code{new()}}{ \subsection{Usage}{ \if{html}{\out{
}}\preformatted{Router$new(path)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{path}}{The base path of the router.} } \if{html}{\out{
}} } \subsection{Details}{ Define the base route. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-Router-print}{}}} \subsection{Method \code{print()}}{ \subsection{Usage}{ \if{html}{\out{
}}\preformatted{Router$print()}\if{html}{\out{
}} } \subsection{Details}{ Print } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-Router-clone}{}}} \subsection{Method \code{clone()}}{ The objects of this class are cloneable with this method. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{Router$clone(deep = FALSE)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{deep}}{Whether to make a deep clone.} } \if{html}{\out{
}} } } }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cookie.R
\name{default_cookie_parser}
\alias{default_cookie_parser}
\title{Cookie Parser}
\usage{
default_cookie_parser(req)
}
\arguments{
\item{req}{A \link{Request}.}
}
\value{
A \code{list} of key value pairs or cookie values.
}
\description{
Parses the cookie string.
}
\examples{
if (interactive()) {
  library(ambiorix)

  #' Handle GET at '/greet'
  #'
  #' @export
  say_hello <- function(req, res) {
    cookies <- default_cookie_parser(req)
    print(cookies)

    res$send("hello there!")
  }

  app <- Ambiorix$new()
  app$get("/greet", say_hello)
  app$start()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{Response}
\alias{Response}
\title{Response}
\value{
A Response object.
}
\description{
Response class to generate responses sent from the server.
}
\examples{
if (interactive()) {
  library(ambiorix)

  app <- Ambiorix$new()

  app$get("/", function(req, res) {
    # print(res)
    res$send("Using {ambiorix}!")
  })

  app$start()
}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{status}}{Status of the response, defaults to \code{200L}.}

\item{\code{headers}}{Named list of headers.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Response-set_status}{\code{Response$set_status()}}
\item \href{#method-Response-send}{\code{Response$send()}}
\item \href{#method-Response-sendf}{\code{Response$sendf()}}
\item \href{#method-Response-text}{\code{Response$text()}}
\item \href{#method-Response-send_file}{\code{Response$send_file()}}
\item \href{#method-Response-redirect}{\code{Response$redirect()}}
\item \href{#method-Response-render}{\code{Response$render()}}
\item \href{#method-Response-json}{\code{Response$json()}}
\item \href{#method-Response-csv}{\code{Response$csv()}}
\item \href{#method-Response-tsv}{\code{Response$tsv()}}
\item \href{#method-Response-htmlwidget}{\code{Response$htmlwidget()}}
\item \href{#method-Response-md}{\code{Response$md()}}
\item \href{#method-Response-png}{\code{Response$png()}}
\item \href{#method-Response-jpeg}{\code{Response$jpeg()}}
\item \href{#method-Response-image}{\code{Response$image()}}
\item \href{#method-Response-ggplot2}{\code{Response$ggplot2()}}
\item \href{#method-Response-print}{\code{Response$print()}}
\item \href{#method-Response-header}{\code{Response$header()}}
\item \href{#method-Response-header_content_json}{\code{Response$header_content_json()}}
\item \href{#method-Response-header_content_html}{\code{Response$header_content_html()}}
\item \href{#method-Response-header_content_plain}{\code{Response$header_content_plain()}}
\item \href{#method-Response-header_content_csv}{\code{Response$header_content_csv()}}
\item \href{#method-Response-header_content_tsv}{\code{Response$header_content_tsv()}}
\item \href{#method-Response-get_headers}{\code{Response$get_headers()}}
\item \href{#method-Response-get_header}{\code{Response$get_header()}}
\item \href{#method-Response-set_headers}{\code{Response$set_headers()}}
\item \href{#method-Response-pre_render_hook}{\code{Response$pre_render_hook()}}
\item \href{#method-Response-post_render_hook}{\code{Response$post_render_hook()}}
\item \href{#method-Response-cookie}{\code{Response$cookie()}}
\item \href{#method-Response-clear_cookie}{\code{Response$clear_cookie()}}
\item \href{#method-Response-clone}{\code{Response$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-set_status"></a>}}
\if{latex}{\out{\hypertarget{method-Response-set_status}{}}}
\subsection{Method \code{set_status()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$set_status(status)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{status}}{An integer defining the status.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Set the status of the response.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-send"></a>}}
\if{latex}{\out{\hypertarget{method-Response-send}{}}}
\subsection{Method \code{send()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$send(body)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{body}}{Body of the response.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a plain HTML response.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-sendf"></a>}}
\if{latex}{\out{\hypertarget{method-Response-sendf}{}}}
\subsection{Method \code{sendf()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$sendf(body, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{body}}{Body of the response.}

\item{\code{...}}{Passed to \code{...} of \code{sprintf}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a plain HTML response, pre-processed with sprintf.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-text"></a>}}
\if{latex}{\out{\hypertarget{method-Response-text}{}}}
\subsection{Method \code{text()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$text(body)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{body}}{Body of the response.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a plain text response.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-send_file"></a>}}
\if{latex}{\out{\hypertarget{method-Response-send_file}{}}}
\subsection{Method \code{send_file()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$send_file(file)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{File to send.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a file.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-redirect"></a>}}
\if{latex}{\out{\hypertarget{method-Response-redirect}{}}}
\subsection{Method \code{redirect()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$redirect(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Path or URL to redirect to.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Redirect to a path or URL.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-render"></a>}}
\if{latex}{\out{\hypertarget{method-Response-render}{}}}
\subsection{Method \code{render()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$render(file, data = list())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{Template file.}

\item{\code{data}}{List to fill \verb{[\% tags \%]}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Render a template file.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-json"></a>}}
\if{latex}{\out{\hypertarget{method-Response-json}{}}}
\subsection{Method \code{json()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$json(body, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{body}}{Body of the response.}

\item{\code{...}}{Additional named arguments passed to the serialiser.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Render an object as JSON.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-csv"></a>}}
\if{latex}{\out{\hypertarget{method-Response-csv}{}}}
\subsection{Method \code{csv()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$csv(data, name = "data", ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{Data to convert to CSV.}

\item{\code{name}}{Name of the file.}

\item{\code{...}}{Additional arguments passed to \code{\link[readr:format_delim]{readr::format_csv()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Sends a comma separated value file
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-tsv"></a>}}
\if{latex}{\out{\hypertarget{method-Response-tsv}{}}}
\subsection{Method \code{tsv()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$tsv(data, name = "data", ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{Data to convert to CSV.}

\item{\code{name}}{Name of the file.}

\item{\code{...}}{Additional arguments passed to \code{\link[readr:format_delim]{readr::format_tsv()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Sends a tab separated value file
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-htmlwidget"></a>}}
\if{latex}{\out{\hypertarget{method-Response-htmlwidget}{}}}
\subsection{Method \code{htmlwidget()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$htmlwidget(widget, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{widget}}{The widget to use.}

\item{\code{...}}{Additional arguments passed to \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Sends an htmlwidget.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-md"></a>}}
\if{latex}{\out{\hypertarget{method-Response-md}{}}}
\subsection{Method \code{md()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$md(file, data = list())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{Template file.}

\item{\code{data}}{List to fill \verb{[\% tags \%]}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Render a markdown file.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-png"></a>}}
\if{latex}{\out{\hypertarget{method-Response-png}{}}}
\subsection{Method \code{png()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$png(file)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{Path to local file.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a png file
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-jpeg"></a>}}
\if{latex}{\out{\hypertarget{method-Response-jpeg}{}}}
\subsection{Method \code{jpeg()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$jpeg(file)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{Path to local file.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a jpeg file
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-image"></a>}}
\if{latex}{\out{\hypertarget{method-Response-image}{}}}
\subsection{Method \code{image()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$image(file)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{Path to local file.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send an image
Similar to \code{png} and \code{jpeg} methods but guesses correct method
based on file extension.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-ggplot2"></a>}}
\if{latex}{\out{\hypertarget{method-Response-ggplot2}{}}}
\subsection{Method \code{ggplot2()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$ggplot2(plot, ..., type = c("png", "jpeg"))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{plot}}{Ggplot2 plot object.}

\item{\code{...}}{Passed to \code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}}

\item{\code{type}}{Type of image to save.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Ggplot2
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-print"></a>}}
\if{latex}{\out{\hypertarget{method-Response-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$print()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Print
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header}{}}}
\subsection{Method \code{header()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header(name, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{String. Name of the header.}

\item{\code{value}}{Value of the header.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Add headers to the response.
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header_content_json"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header_content_json}{}}}
\subsection{Method \code{header_content_json()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header_content_json()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Set Content Type to JSON
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header_content_html"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header_content_html}{}}}
\subsection{Method \code{header_content_html()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header_content_html()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Set Content Type to HTML
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header_content_plain"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header_content_plain}{}}}
\subsection{Method \code{header_content_plain()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header_content_plain()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Set Content Type to Plain Text
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header_content_csv"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header_content_csv}{}}}
\subsection{Method \code{header_content_csv()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header_content_csv()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Set Content Type to CSV
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-header_content_tsv"></a>}}
\if{latex}{\out{\hypertarget{method-Response-header_content_tsv}{}}}
\subsection{Method \code{header_content_tsv()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$header_content_tsv()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Set Content Type to TSV
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-get_headers"></a>}}
\if{latex}{\out{\hypertarget{method-Response-get_headers}{}}}
\subsection{Method \code{get_headers()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$get_headers()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Get headers
Returns the list of headers currently set.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-get_header"></a>}}
\if{latex}{\out{\hypertarget{method-Response-get_header}{}}}
\subsection{Method \code{get_header()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$get_header(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the header to return.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get a header
Returns a single header currently, \code{NULL} if not set.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-set_headers"></a>}}
\if{latex}{\out{\hypertarget{method-Response-set_headers}{}}}
\subsection{Method \code{set_headers()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$set_headers(headers)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{headers}}{A named list of headers to set.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Set headers
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-pre_render_hook"></a>}}
\if{latex}{\out{\hypertarget{method-Response-pre_render_hook}{}}}
\subsection{Method \code{pre_render_hook()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$pre_render_hook(hook)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{hook}}{A function that accepts at least 4 arguments:
\itemize{
\item \code{self}: The \code{Request} class instance.
\item \code{content}: File content a vector of character string,
content of the template.
\item \code{data}: \code{list} passed from \code{render} method.
\item \code{ext}: File extension of the template file.
}

This function is used to add pre-render hooks to the \code{render}
method. The function should return an object of class
\code{responsePreHook} as obtained by \code{\link[=pre_hook]{pre_hook()}}.
This is meant to be used by middlewares to, if necessary,
pre-process rendered data.

Include \code{...} in your \code{hook} to ensure it will handle
potential updates to hooks in the future.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Add a pre render hook.
Runs before the \code{render} method.
}

\subsection{Returns}{
Invisible returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-post_render_hook"></a>}}
\if{latex}{\out{\hypertarget{method-Response-post_render_hook}{}}}
\subsection{Method \code{post_render_hook()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$post_render_hook(hook)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{hook}}{A function to run after the rendering of HTML.
It should accept at least 3 arguments:
\itemize{
\item \code{self}: The \code{Response} class instance.
\item \code{content}: File content a vector of character string,
content of the template.
\item \code{ext}: File extension of the template file.
}

Include \code{...} in your \code{hook} to ensure it will handle
potential updates to hooks in the future.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Post render hook.
}

\subsection{Returns}{
Invisible returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-cookie"></a>}}
\if{latex}{\out{\hypertarget{method-Response-cookie}{}}}
\subsection{Method \code{cookie()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$cookie(
  name,
  value,
  expires = getOption("ambiorix.cookie.expire"),
  max_age = getOption("ambiorix.cookie.maxage"),
  domain = getOption("ambiorix.cookie.domain"),
  path = getOption("ambiorix.cookie.path", "/"),
  secure = getOption("ambiorix.cookie.secure", TRUE),
  http_only = getOption("ambiorix.cookie.httponly", TRUE),
  same_site = getOption("ambiorix.cookie.savesite")
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{String. Name of the cookie.}

\item{\code{value}}{value of the cookie.}

\item{\code{expires}}{Expiry, if an integer assumes it's the number of seconds
from now. Otherwise accepts an object of class \code{POSIXct} or \code{Date}.
If a \code{character} string then it is set as-is and not pre-processed.
If unspecified, the cookie becomes a session cookie. A session finishes
when the client shuts down, after which the session cookie is removed.}

\item{\code{max_age}}{Indicates the number of seconds until the cookie expires.
A zero or negative number will expire the cookie immediately.
If both \code{expires} and \code{max_age} are set, the latter has precedence.}

\item{\code{domain}}{Defines the host to which the cookie will be sent.
If omitted, this attribute defaults to the host of the current document URL,
not including subdomains.}

\item{\code{path}}{Indicates the path that must exist in the requested URL for the
browser to send the Cookie header.}

\item{\code{secure}}{Indicates that the cookie is sent to the server only when a
request is made with the https: scheme (except on localhost), and therefore,
is more resistant to man-in-the-middle attacks.}

\item{\code{http_only}}{Forbids JavaScript from accessing the cookie, for example,
through the document.cookie property.}

\item{\code{same_site}}{Controls whether or not a cookie is sent with cross-origin
requests, providing some protection against cross-site request forgery
attacks (CSRF). Accepts \code{Strict}, \code{Lax}, or \code{None}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Set a cookie
Overwrites existing cookie of the same \code{name}.
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-clear_cookie"></a>}}
\if{latex}{\out{\hypertarget{method-Response-clear_cookie}{}}}
\subsection{Method \code{clear_cookie()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$clear_cookie(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the cookie to clear.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Clear a cookie
Clears the value of a cookie.
}

\subsection{Returns}{
Invisibly returns self.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Response-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Response-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Response$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop.R
\name{stop_all}
\alias{stop_all}
\title{Stop}
\usage{
stop_all()
}
\value{
\code{NULL} (invisibly)
}
\description{
Stop all servers.
}
\examples{
if (interactive()) {
  stop_all()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{content}
\alias{content}
\alias{content_html}
\alias{content_plain}
\alias{content_json}
\alias{content_csv}
\alias{content_tsv}
\alias{content_protobuf}
\title{Content Headers}
\usage{
content_html()

content_plain()

content_json()

content_csv()

content_tsv()

content_protobuf()
}
\value{
Length 1 character vector.
}
\description{
Convenient functions for more readable content type headers.
}
\examples{
list(
 "Content-Type",
 content_json()
)

if(FALSE)
 req$header(
  "Content-Type",
  content_json()
 )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiorix.R
\name{Ambiorix}
\alias{Ambiorix}
\title{Ambiorix}
\value{
An object of class \code{Ambiorix} from which one can
add routes, routers, and run the application.
}
\description{
Web server.
}
\examples{
app <- Ambiorix$new()

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

app$on_stop <- function(){
 cat("Bye!\n")
}

if(interactive())
 app$start()


## ------------------------------------------------
## Method `Ambiorix$listen`
## ------------------------------------------------

app <- Ambiorix$new()

app$listen(3000L)

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()

## ------------------------------------------------
## Method `Ambiorix$set_404`
## ------------------------------------------------

app <- Ambiorix$new()

app$set_404(function(req, res){
 res$send("Nothing found here")
})

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()

## ------------------------------------------------
## Method `Ambiorix$set_error`
## ------------------------------------------------

# my custom error handler:
error_handler <- function(req, res, error) {
  if (!is.null(error)) {
    error_msg <- conditionMessage(error)
    cli::cli_alert_danger("Error: {error_msg}")
  }
  response <- list(
    code = 500L,
    msg = "Uhhmmm... Looks like there's an error from our side :("
  )
  res$
    set_status(500L)$
    json(response)
}

# handler for GET at /whoami:
whoami <- function(req, res) {
  # simulate error (object 'Pikachu' is not defined)
  print(Pikachu)
}

app <- Ambiorix$
  new()$
  set_error(error_handler)$
  get("/whoami", whoami)

if (interactive()) {
  app$start(open = FALSE)
}

## ------------------------------------------------
## Method `Ambiorix$start`
## ------------------------------------------------

app <- Ambiorix$new()

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start(port = 3000L)

## ------------------------------------------------
## Method `Ambiorix$serialiser`
## ------------------------------------------------

app <- Ambiorix$new()

app$serialiser(function(data, ...){
 jsonlite::toJSON(x, ..., pretty = TRUE)
})

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()
}
\section{Super class}{
\code{\link[ambiorix:Routing]{ambiorix::Routing}} -> \code{Ambiorix}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{not_found}}{404 Response, must be a handler function that accepts the request and the response, by default uses \code{\link[=response_404]{response_404()}}.}

\item{\code{error}}{500 response when the route errors, must a handler function that accepts the request and the response, by default uses \code{\link[=response_500]{response_500()}}.}

\item{\code{on_stop}}{Callback function to run when the app stops, takes no argument.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{port}}{Port to run the application.}

\item{\code{host}}{Host to run the application.}

\item{\code{limit}}{Max body size, defaults to \code{5 * 1024 * 1024}.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Ambiorix-new}{\code{Ambiorix$new()}}
\item \href{#method-Ambiorix-cache_templates}{\code{Ambiorix$cache_templates()}}
\item \href{#method-Ambiorix-listen}{\code{Ambiorix$listen()}}
\item \href{#method-Ambiorix-set_404}{\code{Ambiorix$set_404()}}
\item \href{#method-Ambiorix-set_error}{\code{Ambiorix$set_error()}}
\item \href{#method-Ambiorix-static}{\code{Ambiorix$static()}}
\item \href{#method-Ambiorix-start}{\code{Ambiorix$start()}}
\item \href{#method-Ambiorix-serialiser}{\code{Ambiorix$serialiser()}}
\item \href{#method-Ambiorix-stop}{\code{Ambiorix$stop()}}
\item \href{#method-Ambiorix-print}{\code{Ambiorix$print()}}
\item \href{#method-Ambiorix-clone}{\code{Ambiorix$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="engine"><a href='../../ambiorix/html/Routing.html#method-Routing-engine'><code>ambiorix::Routing$engine()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="get_middleware"><a href='../../ambiorix/html/Routing.html#method-Routing-get_middleware'><code>ambiorix::Routing$get_middleware()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="get_params"><a href='../../ambiorix/html/Routing.html#method-Routing-get_params'><code>ambiorix::Routing$get_params()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="get_receivers"><a href='../../ambiorix/html/Routing.html#method-Routing-get_receivers'><code>ambiorix::Routing$get_receivers()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="get_routes"><a href='../../ambiorix/html/Routing.html#method-Routing-get_routes'><code>ambiorix::Routing$get_routes()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="param"><a href='../../ambiorix/html/Routing.html#method-Routing-param'><code>ambiorix::Routing$param()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="prepare"><a href='../../ambiorix/html/Routing.html#method-Routing-prepare'><code>ambiorix::Routing$prepare()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="receive"><a href='../../ambiorix/html/Routing.html#method-Routing-receive'><code>ambiorix::Routing$receive()</code></a></span></li>
<li><span class="pkg-link" data-pkg="ambiorix" data-topic="Routing" data-id="use"><a href='../../ambiorix/html/Routing.html#method-Routing-use'><code>ambiorix::Routing$use()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-new"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$new(
  host = getOption("ambiorix.host", "0.0.0.0"),
  port = getOption("ambiorix.port", NULL),
  log = getOption("ambiorix.logger", TRUE)
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{host}}{A string defining the host.}

\item{\code{port}}{Integer defining the port, defaults to \code{ambiorix.port} option: uses a random port if \code{NULL}.}

\item{\code{log}}{Whether to generate a log of events.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Define the webserver.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-cache_templates"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-cache_templates}{}}}
\subsection{Method \code{cache_templates()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$cache_templates()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Cache templates in memory instead of reading
them from disk.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-listen"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-listen}{}}}
\subsection{Method \code{listen()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$listen(port)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{port}}{Port number.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Specifies the port to listen on.
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$listen(3000L)

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-set_404"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-set_404}{}}}
\subsection{Method \code{set_404()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$set_404(handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handler}}{Function that accepts the request and returns an object
describing an httpuv response, e.g.: \code{\link[=response]{response()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Sets the 404 page.
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$set_404(function(req, res){
 res$send("Nothing found here")
})

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-set_error"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-set_error}{}}}
\subsection{Method \code{set_error()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$set_error(handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handler}}{Function that accepts a request, response and an error object.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Sets the error handler.
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{# my custom error handler:
error_handler <- function(req, res, error) {
  if (!is.null(error)) {
    error_msg <- conditionMessage(error)
    cli::cli_alert_danger("Error: {error_msg}")
  }
  response <- list(
    code = 500L,
    msg = "Uhhmmm... Looks like there's an error from our side :("
  )
  res$
    set_status(500L)$
    json(response)
}

# handler for GET at /whoami:
whoami <- function(req, res) {
  # simulate error (object 'Pikachu' is not defined)
  print(Pikachu)
}

app <- Ambiorix$
  new()$
  set_error(error_handler)$
  get("/whoami", whoami)

if (interactive()) {
  app$start(open = FALSE)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-static"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-static}{}}}
\subsection{Method \code{static()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$static(path, uri = "www")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Local path to directory of assets.}

\item{\code{uri}}{URL path where the directory will be available.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Static directories
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-start"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-start}{}}}
\subsection{Method \code{start()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$start(port = NULL, host = NULL, open = interactive())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{port}}{Integer defining the port, defaults to \code{ambiorix.port} option: uses a random port if \code{NULL}.}

\item{\code{host}}{A string defining the host.}

\item{\code{open}}{Whether to open the app the browser.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Start
Start the webserver.
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start(port = 3000L)
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-serialiser"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-serialiser}{}}}
\subsection{Method \code{serialiser()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$serialiser(handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handler}}{Function to use to serialise.
This function should accept two arguments: the object to serialise and \code{...}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Define Serialiser
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$serialiser(function(data, ...){
 jsonlite::toJSON(x, ..., pretty = TRUE)
})

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

if(interactive())
 app$start()
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-stop"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-stop}{}}}
\subsection{Method \code{stop()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$stop()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Stop
Stop the webserver.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-print"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$print()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Print
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ambiorix-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Ambiorix-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Ambiorix$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{as_renderer}
\alias{as_renderer}
\title{Create a Renderer}
\usage{
as_renderer(fn)
}
\arguments{
\item{fn}{A function that accepts two arguments,
the full path to the \code{file} to render, and the
\code{data} to render.}
}
\value{
A renderer function.
}
\description{
Create a custom renderer.
}
\examples{
if (interactive()) {
  fn <- function(path, data) {
    # ...
  }

  as_renderer(fn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/websocket.R
\name{Websocket}
\alias{Websocket}
\title{Websocket}
\value{
A Websocket object.
}
\description{
Handle websocket messages.
}
\examples{
# create an Ambiorix app with websocket support:
if (interactive()) {
  library(ambiorix)

  home_get <- function(req, res) {
    res$send("hello, world!")
  }

  greeting_ws_handler <- function(msg, ws) {
    cat("Received message:", "\n")
    print(msg)
    ws$send("greeting", "Hello from the server!")
  }

  app <- Ambiorix$new(port = 8080)
  app$get("/", home_get)
  app$receive("greeting", greeting_ws_handler)
  app$start()
}

# create websocket client from another R session:
if (interactive()) {
  client <- websocket::WebSocket$new("ws://127.0.0.1:8080", autoConnect = FALSE)

  client$onOpen(function(event) {
    cat("Connection opened\n")

    msg <- list(
      isAmbiorix = TRUE, # __MUST__ be set!
      name = "greeting",
      message = "Hello from the client!"
    )

    # serialise:
    msg <- yyjsonr::write_json_str(msg, auto_unbox = TRUE)

    client$send(msg)
  })

  client$onMessage(function(event) {
    cat("Received message from server:", event$data, "\n")
  })

  client$connect()
}

}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Websocket-new}{\code{Websocket$new()}}
\item \href{#method-Websocket-send}{\code{Websocket$send()}}
\item \href{#method-Websocket-print}{\code{Websocket$print()}}
\item \href{#method-Websocket-clone}{\code{Websocket$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Websocket-new"></a>}}
\if{latex}{\out{\hypertarget{method-Websocket-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Websocket$new(ws)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ws}}{The websocket}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Constructor
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Websocket-send"></a>}}
\if{latex}{\out{\hypertarget{method-Websocket-send}{}}}
\subsection{Method \code{send()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Websocket$send(name, message)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name, identifier, of the message.}

\item{\code{message}}{Content of the message, anything that can be
serialised to JSON.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Send a message
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Websocket-print"></a>}}
\if{latex}{\out{\hypertarget{method-Websocket-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Websocket$print()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Print
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Websocket-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Websocket-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Websocket$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/routing.R
\name{routing-http-methods}
\alias{routing-http-methods}
\title{Routing HTTP Methods}
\arguments{
\item{path}{String. Route to listen to; use \code{:} to define a parameter (e.g.
\code{"/hello/:name"}).}

\item{handler}{Function that accepts the request and response objects and
returns an httpuv response (e.g. \code{\link[=response]{response()}}). Handlers can return the result
of helper functions such as \code{Response$text()}, \code{Response$json()}, or the
output of any renderer.}

\item{error}{Optional handler invoked if the route raises an error; receives
the request, response, and the error condition.}
}
\value{
The routing object invisibly so calls can be chained.
}
\description{
Register route handlers for HTTP verbs on a \code{\link{Routing}} instance.
}
\details{
The routing helpers provide a fluent API for attaching handlers to HTTP
methods. Each helper shares the same signature and behaviour.
\subsection{Supported helpers}{
\itemize{
\item \code{get()}: Respond to HTTP \code{GET} requests.
\item \code{post()}: Respond to HTTP \code{POST} requests.
\item \code{put()}: Respond to HTTP \code{PUT} requests.
\item \code{patch()}: Respond to HTTP \code{PATCH} requests.
\item \code{delete()}: Respond to HTTP \code{DELETE} requests.
\item \code{options()}: Respond to HTTP \code{OPTIONS} requests.
\item \code{all()}: Respond to every method above.
}
}
}
\examples{
app <- Ambiorix$new()

app$get("/", function(req, res) {
  res$text("Hello, world!")
})

app$post("/echo", function(req, res) {
  res$json(list(received = req$body))
})

app$all("/health", function(req, res) {
  res$json(list(status = "ok"))
})

}
\seealso{
\code{\link{Routing}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import.R
\name{import}
\alias{import}
\title{Import Files}
\usage{
import(...)
}
\arguments{
\item{...}{Directory from which to import \code{.R} or \code{.r} files.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Import all R-files in a directory.
}
\examples{
if (interactive()) {
  import("views")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request.R
\name{Request}
\alias{Request}
\title{Request}
\value{
A Request object.
}
\description{
A request.
}
\examples{
if (interactive()) {
  library(ambiorix)

  app <- Ambiorix$new()

  app$get("/", function(req, res) {
    print(req)
    res$send("Using {ambiorix}!")
  })

  app$start()
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{HEADERS}}{Headers from the request.}

\item{\code{HTTP_ACCEPT}}{Content types to accept.}

\item{\code{HTTP_ACCEPT_ENCODING}}{Encoding of the request.}

\item{\code{HTTP_ACCEPT_LANGUAGE}}{Language of the request.}

\item{\code{HTTP_CACHE_CONTROL}}{Directorives for the cache (case-insensitive).}

\item{\code{HTTP_CONNECTION}}{Controls whether the network connection stays open after the current transaction finishes.}

\item{\code{HTTP_COOKIE}}{Cookie data.}

\item{\code{HTTP_HOST}}{Host making the request.}

\item{\code{HTTP_SEC_FETCH_DEST}}{Indicates the request's destination. That is the initiator of the original fetch request, which is where (and how) the fetched data will be used.}

\item{\code{HTTP_SEC_FETCH_MODE}}{Indicates mode of the request.}

\item{\code{HTTP_SEC_FETCH_SITE}}{Indicates the relationship between a request initiator's origin and the origin of the requested resource.}

\item{\code{HTTP_SEC_FETCH_USER}}{Only sent for requests initiated by user activation, and its value will always be \code{?1}.}

\item{\code{HTTP_UPGRADE_INSECURE_REQUESTS}}{Signals that server supports upgrade.}

\item{\code{HTTP_USER_AGENT}}{User agent.}

\item{\code{SERVER_NAME}}{Name of the server.}

\item{\code{httpuv.version}}{Version of httpuv.}

\item{\code{PATH_INFO}}{Path of the request.}

\item{\code{QUERY_STRING}}{Query string of the request.}

\item{\code{REMOTE_ADDR}}{Remote address.}

\item{\code{REMOTE_PORT}}{Remote port.}

\item{\code{REQUEST_METHOD}}{Method of the request, e.g.: \code{GET}.}

\item{\code{rook.errors}}{Errors from rook.}

\item{\code{rook.input}}{Rook inputs.}

\item{\code{rook.url_scheme}}{Rook url scheme.}

\item{\code{rook.version}}{Rook version.}

\item{\code{SCRIPT_NAME}}{The initial portion of the request URL's "path" that corresponds to the application object, so that the application knows its virtual "location".  #' @field SERVER_NAME Server name.}

\item{\code{SERVER_PORT}}{Server port}

\item{\code{CONTENT_LENGTH}}{Size of the message body.}

\item{\code{CONTENT_TYPE}}{Type of content of the request.}

\item{\code{HTTP_REFERER}}{Contains an absolute or partial address of the page that makes the request.}

\item{\code{body}}{Request, an environment.}

\item{\code{query}}{Parsed \code{QUERY_STRING}, \code{list}.}

\item{\code{params}}{A \code{list} of parameters.}

\item{\code{cookie}}{Parsed \code{HTTP_COOKIE}.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Request-new}{\code{Request$new()}}
\item \href{#method-Request-print}{\code{Request$print()}}
\item \href{#method-Request-get_header}{\code{Request$get_header()}}
\item \href{#method-Request-parse_multipart}{\code{Request$parse_multipart()}}
\item \href{#method-Request-parse_json}{\code{Request$parse_json()}}
\item \href{#method-Request-clone}{\code{Request$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-new"></a>}}
\if{latex}{\out{\hypertarget{method-Request-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$new(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{Original request (environment).}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Constructor
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-print"></a>}}
\if{latex}{\out{\hypertarget{method-Request-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$print()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Print
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-get_header"></a>}}
\if{latex}{\out{\hypertarget{method-Request-get_header}{}}}
\subsection{Method \code{get_header()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$get_header(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the header}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get Header
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-parse_multipart"></a>}}
\if{latex}{\out{\hypertarget{method-Request-parse_multipart}{}}}
\subsection{Method \code{parse_multipart()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$parse_multipart()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Parse Multipart encoded data
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-parse_json"></a>}}
\if{latex}{\out{\hypertarget{method-Request-parse_json}{}}}
\subsection{Method \code{parse_json()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$parse_json(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Arguments passed to \code{\link[=parse_json]{parse_json()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Parse JSON encoded data
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Request-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Request-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Request$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request.R
\name{mockRequest}
\alias{mockRequest}
\title{Mock Request}
\usage{
mockRequest(cookie = "", query = "", path = "/")
}
\arguments{
\item{cookie}{Cookie string.}

\item{query}{Query string.}

\item{path}{Path string.}
}
\value{
A \code{Request} object.
}
\description{
Mock a request, used for tests.
}
\examples{
mockRequest()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/routing.R
\name{Routing}
\alias{Routing}
\title{Core Routing Class}
\value{
A Routing object.
}
\description{
Core routing class.
Do not use directly, see \link{Ambiorix}, and \link{Router}.
}
\section{HTTP methods}{

See \link{routing-http-methods} for the full argument reference. The routing
instance exposes helpers for common HTTP verbs; they are registered when the
object is initialised and share the same signature.
\itemize{
\item \code{get()}, \code{put()}, \code{patch()}, \code{delete()}, \code{post()}, \code{options()} register a
handler for the single corresponding HTTP verb; see
\link{routing-http-methods}.
\item \code{all()} registers a handler that responds to \code{GET}, \code{POST}, \code{PUT},
\code{DELETE}, and \code{PATCH}; see \link{routing-http-methods}.
}
}

\examples{

## ------------------------------------------------
## Method `Routing$param`
## ------------------------------------------------

app <- Ambiorix$new()

app$get("/", function(req,res){
 res$send("Hello!")
})

app$param("person", function(req, res, value, name){
 if(value == "notWanted"){
  res$status <- 403L
  res$send("This is the end.")
 }

 # continue processing the request...
})

app$get("/hi/:person", function(req,res){
 res$sendf("Hi! \%s", req$params$person)
})
app$get("/info/:person", function(req,res){
 res$sendf("Here is all your info, \%s", req$params$person)
})
if(interactive())
 app$start()

## ------------------------------------------------
## Method `Routing$receive`
## ------------------------------------------------

app <- Ambiorix$new()

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

app$receive("hello", function(msg, ws){
 print(msg) # print msg received

 # send a message back
 ws$send("hello", "Hello back! (sent from R)")
})

if(interactive())
 app$start()
}
\seealso{
\link{routing-http-methods}
}
\keyword{export}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{error}}{Error handler.}

\item{\code{get}}{Register a route handler for HTTP GET requests. See
\link{routing-http-methods}.}

\item{\code{put}}{Register a route handler for HTTP PUT requests. See
\link{routing-http-methods}.}

\item{\code{patch}}{Register a route handler for HTTP PATCH requests. See
\link{routing-http-methods}.}

\item{\code{delete}}{Register a route handler for HTTP DELETE requests. See
\link{routing-http-methods}.}

\item{\code{post}}{Register a route handler for HTTP POST requests. See
\link{routing-http-methods}.}

\item{\code{options}}{Register a route handler for HTTP OPTIONS requests. See
\link{routing-http-methods}.}

\item{\code{all}}{Register a route handler that responds to every HTTP verb used by
Ambiorix. See \link{routing-http-methods}.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{basepath}}{Basepath, read-only.}

\item{\code{websocket}}{Websocket handler.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Routing-new}{\code{Routing$new()}}
\item \href{#method-Routing-param}{\code{Routing$param()}}
\item \href{#method-Routing-receive}{\code{Routing$receive()}}
\item \href{#method-Routing-print}{\code{Routing$print()}}
\item \href{#method-Routing-engine}{\code{Routing$engine()}}
\item \href{#method-Routing-use}{\code{Routing$use()}}
\item \href{#method-Routing-get_routes}{\code{Routing$get_routes()}}
\item \href{#method-Routing-get_params}{\code{Routing$get_params()}}
\item \href{#method-Routing-get_receivers}{\code{Routing$get_receivers()}}
\item \href{#method-Routing-get_middleware}{\code{Routing$get_middleware()}}
\item \href{#method-Routing-prepare}{\code{Routing$prepare()}}
\item \href{#method-Routing-clone}{\code{Routing$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-new"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$new(path = "")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Prefix path.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Initialise
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-param"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-param}{}}}
\subsection{Method \code{param()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$param(name, handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the parameter}

\item{\code{handler}}{Function that accepts the request, response, parameter value and the parameter name.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
PARAM Method
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$get("/", function(req,res){
 res$send("Hello!")
})

app$param("person", function(req, res, value, name){
 if(value == "notWanted"){
  res$status <- 403L
  res$send("This is the end.")
 }

 # continue processing the request...
})

app$get("/hi/:person", function(req,res){
 res$sendf("Hi! \%s", req$params$person)
})
app$get("/info/:person", function(req,res){
 res$sendf("Here is all your info, \%s", req$params$person)
})
if(interactive())
 app$start()
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-receive"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-receive}{}}}
\subsection{Method \code{receive()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$receive(name, handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of message.}

\item{\code{handler}}{Function to run when message is received.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Receive Websocket Message
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{app <- Ambiorix$new()

app$get("/", function(req, res){
 res$send("Using {ambiorix}!")
})

app$receive("hello", function(msg, ws){
 print(msg) # print msg received

 # send a message back
 ws$send("hello", "Hello back! (sent from R)")
})

if(interactive())
 app$start()
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-print"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$print()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Print
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-engine"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-engine}{}}}
\subsection{Method \code{engine()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$engine(engine)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{engine}}{Engine function.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Engine to use for rendering templates.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-use"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-use}{}}}
\subsection{Method \code{use()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$use(use)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{use}}{Either a router as returned by \link{Router}, a function to use as middleware,
or a \code{list} of functions.
If a function is passed, it must accept two arguments (the request, and the response):
this function will be executed every time the server receives a request.
\emph{Middleware may but does not have to return a response, unlike other methods such as \code{get}}
Note that multiple routers and middlewares can be used.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Use a router or middleware
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-get_routes"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-get_routes}{}}}
\subsection{Method \code{get_routes()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$get_routes(routes = list(), parent = "")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{routes}}{Existing list of routes.}

\item{\code{parent}}{Parent path.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get the routes
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-get_params"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-get_params}{}}}
\subsection{Method \code{get_params()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$get_params(params = list(), parent = "")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{params}}{Existing list of parameter middlewares.}

\item{\code{parent}}{Parent path.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get the parameter middlewares
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-get_receivers"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-get_receivers}{}}}
\subsection{Method \code{get_receivers()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$get_receivers(receivers = list())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{receivers}}{Existing list of receivers}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get the websocket receivers
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-get_middleware"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-get_middleware}{}}}
\subsection{Method \code{get_middleware()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$get_middleware(middlewares = list(), parent = "")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{middlewares}}{Existing list of middleswares}

\item{\code{parent}}{Parent path}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Get the middleware
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-prepare"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-prepare}{}}}
\subsection{Method \code{prepare()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$prepare()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Prepare routes and decomposes paths
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Routing-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Routing-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Routing$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{new_log}
\alias{new_log}
\title{Logger}
\usage{
new_log(prefix = ">", write = FALSE, file = "ambiorix.log", sep = "")
}
\arguments{
\item{prefix}{String to prefix all log messages.}

\item{write}{Whether to write the log to the \code{file}.}

\item{file}{Name of the file to dump the logs to,
only used if \code{write} is \code{TRUE}.}

\item{sep}{Separator between \code{prefix} and other
flags and messages.}
}
\value{
An R& of class \code{log::Logger}.
}
\description{
Returns a new logger using the \code{log} package.
}
\examples{
log <- new_log()
log$log("Hello world")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cookie.R
\name{as_cookie_preprocessor}
\alias{as_cookie_preprocessor}
\title{Define a Cookie Preprocessor}
\usage{
as_cookie_preprocessor(fn)
}
\arguments{
\item{fn}{A function that accepts the same arguments
as the \code{cookie} method of the \link{Response} class
(name, value, ...), and returns a modified
\code{value}.}
}
\value{
Object of class "cookiePreprocessor".
}
\description{
Identifies a function as a cookie preprocessor.
}
\examples{
func <- function(name, value, ...) {
 sprintf("prefix.\%s", value)
}

prep <- as_cookie_preprocessor(func)

app <- Ambiorix$new()
app$use(prep)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token.R
\name{token_create}
\alias{token_create}
\title{Token}
\usage{
token_create(n = 16L)
}
\arguments{
\item{n}{Number of bytes.}
}
\value{
Length 1 character vector.
}
\description{
Create a token
}
\examples{
token_create()
token_create(n = 32L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{robj}
\alias{robj}
\title{R Object}
\usage{
robj(obj)
}
\arguments{
\item{obj}{R object to treat.}
}
\value{
Object of class "robj".
}
\description{
Treats a data element rendered in a response (\code{res$render}) as
a data object and ultimately uses \code{\link[=dput]{dput()}}.
}
\details{
For instance in a template, \verb{x <- [\% var \%]} will not work with
\code{res$render(data=list(var = "hello"))} because this will be replace
like \code{x <- hello} (missing quote): breaking the template. Using \code{robj} one would
obtain \code{x <- "hello"}.
}
\examples{
robj(1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/websocket.R
\name{websocket_client}
\alias{websocket_client}
\alias{copy_websocket_client}
\alias{get_websocket_client_path}
\alias{get_websocket_clients}
\title{Websocket Client}
\usage{
copy_websocket_client(path)

get_websocket_client_path()

get_websocket_clients()
}
\arguments{
\item{path}{Path to copy the file to.}
}
\value{
\itemize{
\item \code{copy_websocket_client}: String. The new path (invisibly).
\item \code{get_websocket_client_path}: String. The path to the local websocket client.
\item \code{get_websocket_clients}: List. Websocket clients.
}
}
\description{
Handle ambiorix websocket client.
}
\section{Functions}{

\itemize{
\item \code{copy_websocket_client} Copies the websocket client file, useful when
ambiorix was not setup with the ambiorix generator.
\item \code{get_websocket_client_path} Retrieves the full path to the local websocket client.
\item \code{get_websocket_clients} Retrieves clients connected to the server.
}
}

\examples{
chat_ws <- function(msg, ws) {
  lapply(
    X = get_websocket_clients(),
    FUN = function(c) {
      c$send("chat", msg)
    }
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{set_log}
\alias{set_log}
\alias{set_log_info}
\alias{set_log_success}
\alias{set_log_error}
\title{Customise logs}
\usage{
set_log_info(log)

set_log_success(log)

set_log_error(log)
}
\arguments{
\item{log}{An object of class \code{Logger}, see
\link[log:Logger]{log::Logger}.}
}
\value{
The \code{log} object.
}
\description{
Customise the internal logs used by Ambiorix.
}
\examples{
# define custom loggers:
info_logger <- log::Logger$new("INFO")
success_logger <- log::Logger$new("SUCCESS")
error_logger <- log::Logger$new("ERROR")

info_logger$log("This is an info message.")
success_logger$log("This is a success message.")
error_logger$log("This is an error message.")

# set custom loggers for Ambiorix:
set_log_info(info_logger)
set_log_success(success_logger)
set_log_error(error_logger)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forward.R
\name{forward}
\alias{forward}
\title{Forward Method}
\usage{
forward()
}
\value{
An object of class \code{forward}.
}
\description{
Makes it such that the web server skips this method and uses the next one in line instead.
}
\examples{
app <- Ambiorix$new()

app$get("/next", function(req, res){
 forward()
})

app$get("/next", function(req, res){
 res$send("Hello")
})

if(interactive())
 app$start()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{responses}
\alias{responses}
\alias{response}
\alias{response_404}
\alias{response_500}
\title{Plain Responses}
\usage{
response(body, headers = list(), status = 200L)

response_404(
  body = "404: Not found",
  headers = list(`Content-Type` = content_html()),
  status = 404L
)

response_500(
  body = "500: Server Error",
  headers = list(`Content-Type` = content_html()),
  status = 500L
)
}
\arguments{
\item{body}{Body of response.}

\item{headers}{HTTP headers.}

\item{status}{Response status}
}
\value{
An Ambiorix response.
}
\description{
Plain HTTP Responses.
}
\examples{
app <- Ambiorix$new()

# html
app$get("/", function(req, res){
 res$send("hello!")
})

# text
app$get("/text", function(req, res){
 res$text("hello!")
})

if(interactive())
 app$start()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parser.R
\name{parse_form_urlencoded}
\alias{parse_form_urlencoded}
\title{Parse application/x-www-form-urlencoded data}
\usage{
parse_form_urlencoded(req, ...)
}
\arguments{
\item{req}{The request object.}

\item{...}{Additional parameters passed to the parser function.}
}
\value{
A list of parsed form fields, with each key representing a form field name and each value
representing the form field's value.

Named list
}
\description{
This function parses \code{application/x-www-form-urlencoded} data, typically used in form submissions.
}
\details{
\subsection{Overriding Default Parser}{

By default, \code{parse_form_urlencoded()} uses \code{\link[webutils:parse_http]{webutils::parse_http()}}.
You can override this globally by setting the \code{AMBIORIX_FORM_URLENCODED_PARSER} option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(AMBIORIX_FORM_URLENCODED_PARSER = my_other_custom_parser)
}\if{html}{\out{</div>}}

Your custom parser function \emph{MUST} accept the following parameters:
\enumerate{
\item \code{body}: Raw vector containing the form data.
\item \code{...}: Additional optional parameters.
}
}
}
\examples{
if (interactive()) {
  library(ambiorix)
  library(htmltools)
  library(readxl)

  page_links <- function() {
    Map(
      f = function(href, label) {
        tags$a(href = href, label)
      },
      c("/", "/about", "/contact"),
      c("Home", "About", "Contact")
    )
  }

  forms <- function() {
    form1 <- tags$form(
      action = "/url-form-encoded",
      method = "POST",
      enctype = "application/x-www-form-urlencoded",
      tags$h4("form-url-encoded:"),
      tags$label(`for` = "first_name", "First Name"),
      tags$input(id = "first_name", name = "first_name", value = "John"),
      tags$label(`for` = "last_name", "Last Name"),
      tags$input(id = "last_name", name = "last_name", value = "Coene"),
      tags$button(type = "submit", "Submit")
    )

    form2 <- tags$form(
      action = "/multipart-form-data",
      method = "POST",
      enctype = "multipart/form-data",
      tags$h4("multipart/form-data:"),
      tags$label(`for` = "email", "Email"),
      tags$input(id = "email", name = "email", value = "john@mail.com"),
      tags$label(`for` = "framework", "Framework"),
      tags$input(id = "framework", name = "framework", value = "ambiorix"),
      tags$label(`for` = "file", "Upload CSV file"),
      tags$input(type = "file", id = "file", name = "file", accept = ".csv"),
      tags$label(`for` = "file2", "Upload xlsx file"),
      tags$input(type = "file", id = "file2", name = "file2", accept = ".xlsx"),
      tags$button(type = "submit", "Submit")
    )

    tagList(form1, form2)
  }

  home_get <- function(req, res) {
    html <- tagList(
      page_links(),
      tags$h3("hello, world!"),
      forms()
    )

    res$send(html)
  }

  home_post <- function(req, res) {
    body <- parse_json(req)
    # print(body)

    response <- list(
      code = 200L,
      msg = "hello, world"
    )
    res$json(response)
  }

  url_form_encoded_post <- function(req, res) {
    body <- parse_form_urlencoded(req)
    # print(body)

    list_items <- lapply(
      X = names(body),
      FUN = function(nm) {
        tags$li(
          nm,
          ":",
          body[[nm]]
        )
      }
    )
    input_vals <- tags$ul(list_items)

    html <- tagList(
      page_links(),
      tags$h3("Request processed"),
      input_vals
    )

    res$send(html)
  }

  multipart_form_data_post <- function(req, res) {
    body <- parse_multipart(req)

    list_items <- lapply(
      X = names(body),
      FUN = function(nm) {
        field <- body[[nm]]

        # if 'field' is a file, parse it & print on console:
        is_file <- "filename" \%in\% names(field)
        is_csv <- is_file && identical(field[["content_type"]], "text/csv")
        is_xlsx <- is_file &&
          identical(
            field[["content_type"]],
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
          )

        if (is_file) {
          file_path <- tempfile()
          writeBin(object = field$value, con = file_path)
          on.exit(unlink(x = file_path))
        }

        if (is_csv) {
          # print(read.csv(file = file_path))
        }

        if (is_xlsx) {
          # print(readxl::read_xlsx(path = file_path))
        }

        tags$li(
          nm,
          ":",
          if (is_file) "printed on console" else field
        )
      }
    )
    input_vals <- tags$ul(list_items)

    html <- tagList(
      page_links(),
      tags$h3("Request processed"),
      input_vals
    )

    res$send(html)
  }

  about_get <- function(req, res) {
    html <- tagList(
      page_links(),
      tags$h3("About Us")
    )
    res$send(html)
  }

  contact_get <- function(req, res) {
    html <- tagList(
      page_links(),
      tags$h3("Get In Touch!")
    )
    res$send(html)
  }

  app <- Ambiorix$new(port = 5000L)

  app$
    get("/", home_get)$
    post("/", home_post)$
    get("/about", about_get)$
    get("/contact", contact_get)$
    post("/url-form-encoded", url_form_encoded_post)$
    post("/multipart-form-data", multipart_form_data_post)

  app$start()
}
}
\seealso{
\code{\link[=parse_multipart]{parse_multipart()}}, \code{\link[=parse_json]{parse_json()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialise.R
\name{serialise}
\alias{serialise}
\title{Serialise an Object to JSON}
\usage{
serialise(data, ...)
}
\arguments{
\item{data}{Data to serialise.}

\item{...}{Passed to serialiser.}
}
\value{
JSON string.
}
\description{
Serialise an Object to JSON
}
\details{
Ambiorix uses \code{\link[yyjsonr:write_json_str]{yyjsonr::write_json_str()}} by default for serialization.
\subsection{Custom Serialiser}{

To override the default, set the \code{AMBIORIX_SERIALISER} option to a function that accepts:
\itemize{
\item \code{data}: Object to serialise.
\item \code{...}: Additional arguments passed to the function.
}

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{my_serialiser <- function(data, ...) \{
 jsonlite::toJSON(x = data, ...)
\}

options(AMBIORIX_SERIALISER = my_serialiser)
}\if{html}{\out{</div>}}
}
}
\examples{
if (interactive()) {
  # a list:
  response <- list(code = 200L, msg = "hello, world!")

  serialise(response)
  #> {"code":200,"msg":"hello, world"}

  serialise(response, auto_unbox = FALSE)
  #> {"code":[200],"msg":["hello, world"]}

  # data.frame:
  serialise(cars)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{jobj}
\alias{jobj}
\title{JSON Object}
\usage{
jobj(obj)
}
\arguments{
\item{obj}{Object to serialise.}
}
\value{
Object of class "jobj".
}
\description{
Serialises an object to JSON in \code{res$render}.
}
\examples{
if (interactive()) {
  l <- list(a = "hello", b = 2L, c = 3)
  jobj(l)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/route.R
\name{as_path_to_pattern}
\alias{as_path_to_pattern}
\title{Path to pattern}
\usage{
as_path_to_pattern(path)
}
\arguments{
\item{path}{A function that accepts a character vector of length 1
and returns another character vector of length 1.}
}
\value{
Object of class "pathToPattern".
}
\description{
Identify a function as a path to pattern function;
a function that accepts a path and returns a matching pattern.
}
\examples{
fn <- function(path) {
  pattern <- gsub(":([^/]+)", "(\\\\\\\\w+)", path)
  paste0("^", pattern, "$")
}

path_to_pattern <- as_path_to_pattern(fn)

path <- "/dashboard/profile/:user_id"
pattern <- path_to_pattern(path) # "^/dashboard/profile/(\\w+)$"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cookie.R
\name{as_cookie_parser}
\alias{as_cookie_parser}
\title{Define a Cookie Parser}
\usage{
as_cookie_parser(fn)
}
\arguments{
\item{fn}{A function that accepts a single argument,
\code{req} the \link{Request} and returns the parsed cookie string,
generally a \code{list}.
Note that the original cookie string is available on the
\link{Request} at the \code{HTTP_COOKIE} field, get it with:
\code{req$HTTP_COOKIE}}
}
\value{
Object of class "cookieParser".
}
\description{
Identifies a function as a cookie parser (see example).
}
\examples{
func <- function(req) {
 req$HTTP_COOKIE
}

parser <- as_cookie_parser(func)

app <- Ambiorix$new()
app$use(parser)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{pre_hook}
\alias{pre_hook}
\title{Pre Hook Response}
\usage{
pre_hook(content, data)
}
\arguments{
\item{content}{File content, a character vector.}

\item{data}{A list of data passed to \code{glue::glue_data}.}
}
\value{
A response pre-hook.
}
\description{
Pre Hook Response
}
\examples{
my_prh <- function(self, content, data, ext, ...) {
  data$title <- "Mansion"
  pre_hook(content, data)
}

#' Handler for GET at '/'
#'
#' @details Renders the homepage
#' @export
home_get <- function(req, res) {
  res$pre_render_hook(my_prh)
  res$render(
    file = "page.html",
    data = list(
      title = "Home"
    )
  )
}
}
